#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ast
import os
import tarfile
from unittest import mock

from kfp import compiler as kfp_argo_compiler
import pytest
import yaml

from elyra.metadata.metadata import Metadata
from elyra.pipeline.component import Component
from elyra.pipeline.kfp.processor_kfp import KfpPipelineProcessor
from elyra.pipeline.parser import PipelineParser
from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.pipeline import Operation
from elyra.pipeline.pipeline import Pipeline
from elyra.tests.pipeline.test_pipeline_parser import _read_pipeline_resource


@pytest.fixture
def processor(setup_factory_data):
    processor = KfpPipelineProcessor(os.getcwd())
    return processor


@pytest.fixture
def pipeline():
    pipeline_resource = _read_pipeline_resource(
        'resources/sample_pipelines/pipeline_3_node_sample.json')
    return PipelineParser.parse(pipeline_resource)


@pytest.fixture
def sample_metadata():
    return {"api_endpoint": "http://examples.com:31737",
            "cos_endpoint": "http://examples.com:31671",
            "cos_username": "example",
            "cos_password": "example123",
            "cos_bucket": "test",
            "engine": "Argo",
            "tags": []}


def test_fail_get_metadata_configuration_invalid_namespace(processor):
    with pytest.raises(RuntimeError):
        processor._get_metadata_configuration(schemaspace="non_existent_namespace",
                                              name='non_existent_metadata')


def test_generate_dependency_archive(processor):
    pipelines_test_file = processor.root_dir + '/elyra/tests/pipeline/resources/archive/test.ipynb'
    pipeline_dependencies = ['airflow.json']
    correct_filelist = ['test.ipynb', 'airflow.json']
    component_parameters = {
        'filename': pipelines_test_file,
        'dependencies': pipeline_dependencies,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='123e4567-e89b-12d3-a456-426614174000',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    archive_location = processor._generate_dependency_archive(test_operation)

    tar_content = []
    with tarfile.open(archive_location, "r:gz") as tar:
        for tarinfo in tar:
            if tarinfo.isreg():
                print(tarinfo.name)
                tar_content.append(tarinfo.name)

    assert sorted(correct_filelist) == sorted(tar_content)


def test_fail_generate_dependency_archive(processor):
    pipelines_test_file = processor.root_dir + '/elyra/pipeline/tests/resources/archive/test.ipynb'
    pipeline_dependencies = ['non_existent_file.json']
    component_parameters = {
        'filename': pipelines_test_file,
        'dependencies': pipeline_dependencies,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='123e4567-e89b-12d3-a456-426614174000',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    with pytest.raises(Exception):
        processor._generate_dependency_archive(test_operation)


def test_get_dependency_source_dir(processor):
    pipelines_test_file = 'elyra/pipeline/tests/resources/archive/test.ipynb'
    processor.root_dir = '/this/is/an/abs/path/'
    correct_filepath = '/this/is/an/abs/path/elyra/pipeline/tests/resources/archive'
    component_parameters = {
        'filename': pipelines_test_file,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='123e4567-e89b-12d3-a456-426614174000',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    filepath = processor._get_dependency_source_dir(test_operation)

    assert filepath == correct_filepath


def test_get_dependency_archive_name(processor):
    pipelines_test_file = 'elyra/pipeline/tests/resources/archive/test.ipynb'
    correct_filename = 'test-this-is-a-test-id.tar.gz'
    component_parameters = {
        'filename': pipelines_test_file,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='this-is-a-test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    filename = processor._get_dependency_archive_name(test_operation)

    assert filename == correct_filename


def test_collect_envs(processor):
    pipelines_test_file = 'elyra/pipeline/tests/resources/archive/test.ipynb'

    # add system-owned envs with bogus values to ensure they get set to system-derived values,
    # and include some user-provided edge cases
    operation_envs = ['ELYRA_RUNTIME_ENV="bogus_runtime"',
                      'ELYRA_ENABLE_PIPELINE_INFO="bogus_pipeline"',
                      'ELYRA_WRITABLE_CONTAINER_DIR=',  # simulate operation reference in pipeline
                      'AWS_ACCESS_KEY_ID="bogus_key"',
                      'AWS_SECRET_ACCESS_KEY="bogus_secret"',
                      'USER_EMPTY_VALUE=  ',
                      'USER_TWO_EQUALS=KEY=value',
                      'USER_NO_VALUE=']

    component_parameters = {
        'filename': pipelines_test_file,
        'env_vars': operation_envs,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='this-is-a-test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    envs = processor._collect_envs(test_operation, cos_secret=None, cos_username='Alice', cos_password='secret')

    assert envs['ELYRA_RUNTIME_ENV'] == 'kfp'
    assert envs['AWS_ACCESS_KEY_ID'] == 'Alice'
    assert envs['AWS_SECRET_ACCESS_KEY'] == 'secret'
    assert envs['ELYRA_ENABLE_PIPELINE_INFO'] == 'True'
    assert envs['ELYRA_WRITABLE_CONTAINER_DIR'] == '/tmp'
    assert envs['USER_EMPTY_VALUE'] == '  '
    assert envs['USER_TWO_EQUALS'] == 'KEY=value'
    assert 'USER_NO_VALUE' not in envs


# Repeat with non-None secret - ensure user and password envs are not present, but others are
    envs = processor._collect_envs(test_operation, cos_secret='secret', cos_username='Alice', cos_password='secret')

    assert envs['ELYRA_RUNTIME_ENV'] == 'kfp'
    assert 'AWS_ACCESS_KEY_ID' not in envs
    assert 'AWS_SECRET_ACCESS_KEY' not in envs
    assert envs['ELYRA_ENABLE_PIPELINE_INFO'] == 'True'
    assert envs['ELYRA_WRITABLE_CONTAINER_DIR'] == '/tmp'
    assert envs['USER_EMPTY_VALUE'] == '  '
    assert envs['USER_TWO_EQUALS'] == 'KEY=value'
    assert 'USER_NO_VALUE' not in envs


def test_process_list_value_function(processor):
    # Test values that will be successfully converted to list
    assert processor._process_list_value("") == []
    assert processor._process_list_value(None) == []
    assert processor._process_list_value("[]") == []
    assert processor._process_list_value("None") == []
    assert processor._process_list_value("['elem1']") == ["elem1"]
    assert processor._process_list_value("['elem1', 'elem2', 'elem3']") == ["elem1", "elem2", "elem3"]
    assert processor._process_list_value("  ['elem1',   'elem2' , 'elem3']  ") == ["elem1", "elem2", "elem3"]
    assert processor._process_list_value("[1, 2]") == [1, 2]
    assert processor._process_list_value("[True, False, True]") == [True, False, True]
    assert processor._process_list_value("[{'obj': 'val', 'obj2': 'val2'}, {}]") == [{'obj': 'val', 'obj2': 'val2'}, {}]

    # Test values that will not be successfully converted to list
    assert processor._process_list_value("[[]") == "[[]"
    assert processor._process_list_value("[elem1, elem2]") == "[elem1, elem2]"
    assert processor._process_list_value("elem1, elem2") == "elem1, elem2"
    assert processor._process_list_value("  elem1, elem2  ") == "elem1, elem2"
    assert processor._process_list_value("'elem1', 'elem2'") == "'elem1', 'elem2'"


def test_process_dictionary_value_function(processor):
    # Test values that will be successfully converted to dictionary
    assert processor._process_dictionary_value("") == {}
    assert processor._process_dictionary_value(None) == {}
    assert processor._process_dictionary_value("{}") == {}
    assert processor._process_dictionary_value("None") == {}
    assert processor._process_dictionary_value("{'key': 'value'}") == {"key": "value"}

    dict_as_str = "{'key1': 'value', 'key2': 'value'}"
    assert processor._process_dictionary_value(dict_as_str) == {"key1": "value", "key2": "value"}

    dict_as_str = "  {  'key1': 'value'  , 'key2'  : 'value'}  "
    assert processor._process_dictionary_value(dict_as_str) == {"key1": "value", "key2": "value"}

    dict_as_str = "{'key1': [1, 2, 3], 'key2': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == {"key1": [1, 2, 3], "key2": ["elem1", "elem2"]}

    dict_as_str = "{'key1': 2, 'key2': 'value', 'key3': True, 'key4': None, 'key5': [1, 2, 3]}"
    expected_value = {
        "key1": 2,
        "key2": "value",
        "key3": True,
        "key4": None,
        "key5": [1, 2, 3]
    }
    assert processor._process_dictionary_value(dict_as_str) == expected_value

    dict_as_str = "{'key1': {'key2': 2, 'key3': 3, 'key4': 4}, 'key5': {}}"
    expected_value = {
        "key1": {
            "key2": 2,
            "key3": 3,
            "key4": 4,
        },
        "key5": {}
    }
    assert processor._process_dictionary_value(dict_as_str) == expected_value

    # Test values that will not be successfully converted to dictionary
    assert processor._process_dictionary_value("{{}") == "{{}"
    assert processor._process_dictionary_value("{key1: value, key2: value}") == "{key1: value, key2: value}"
    assert processor._process_dictionary_value("  { key1: value, key2: value }  ") == "{ key1: value, key2: value }"
    assert processor._process_dictionary_value("key1: value, key2: value") == "key1: value, key2: value"
    assert processor._process_dictionary_value("{'key1': true}") == "{'key1': true}"
    assert processor._process_dictionary_value("{'key': null}") == "{'key': null}"

    dict_as_str = "{'key1': [elem1, elem2, elem3], 'key2': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == dict_as_str

    dict_as_str = "{'key1': {key2: 2}, 'key3': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == dict_as_str

    dict_as_str = "{'key1': {key2: 2}, 'key3': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == dict_as_str


def test_processing_url_runtime_specific_component(monkeypatch, processor, sample_metadata, tmpdir):
    # Assign test resource location
    url = 'https://raw.githubusercontent.com/elyra-ai/elyra/master/' \
          'elyra/tests/pipeline/resources/components/filter_text.yaml'

    # Instantiate a url-based component
    component_id = 'filter-text'
    component = Component(id=component_id,
                          name="Filter text",
                          description="",
                          op="filter-text",
                          location_type="url",
                          location=url,
                          properties=[],
                          categories=[])

    # Replace cached component registry with single url-based component for testing
    processor._component_registry._cached_components = {component_id: component}

    # Construct hypothetical operation for component
    operation_name = "Filter text test"
    operation_params = {
        "text": "path/to/text.txt",
        "pattern": "hello"
    }
    operation = Operation(id='filter-text-id',
                          type='execution_node',
                          classifier=component_id,
                          name=operation_name,
                          parent_operation_ids=[],
                          component_params=operation_params)

    # Build a mock runtime config for use in _cc_pipeline
    mocked_runtime = Metadata(name="test-metadata",
                              display_name="test",
                              schema_name="airflow",
                              metadata=sample_metadata)

    mocked_func = mock.Mock(return_value="default", side_effect=[mocked_runtime, sample_metadata])
    monkeypatch.setattr(processor, "_get_metadata_configuration", mocked_func)

    # Construct single-operation pipeline
    pipeline = Pipeline(id='pipeline-id',
                        name='kfp_test',
                        runtime='kfp',
                        runtime_config='test',
                        source='filter_text.pipeline')
    pipeline.operations[operation.id] = operation

    # Establish path and function to construct pipeline
    pipeline_path = os.path.join(tmpdir, 'kfp_test.yaml')
    constructed_pipeline_function = lambda: processor._cc_pipeline(pipeline=pipeline, pipeline_name='test_pipeline')

    # TODO Check against both argo and tekton compilations
    # Compile pipeline and save into pipeline_path
    kfp_argo_compiler.Compiler().compile(constructed_pipeline_function, pipeline_path)

    # Read contents of pipeline YAML
    with open(pipeline_path) as f:
        pipeline_yaml = yaml.safe_load(f.read())

    # Check the pipeline file contents for correctness
    pipeline_template = pipeline_yaml['spec']['templates'][0]
    assert pipeline_template['metadata']['annotations']['pipelines.kubeflow.org/task_display_name'] == operation_name
    assert pipeline_template['inputs']['artifacts'][0]['raw']['data'] == operation_params['text']

    component_ref = pipeline_template['metadata']['annotations']['pipelines.kubeflow.org/component_ref']
    component_ref = ast.literal_eval(component_ref)
    assert component_ref['url'] == url


def test_processing_filename_runtime_specific_component(monkeypatch, processor, sample_metadata, tmpdir):
    # Assign test resource location
    relative_path = "kfp/filter_text_using_shell_and_grep.yaml"

    # Instantiate a file-based component
    component_id = "filter-text"
    component = Component(id=component_id,
                          name="Filter text",
                          description="",
                          op="filter-text",
                          location_type="filename",
                          location=relative_path,
                          properties=[],
                          categories=[])

    # Replace cached component registry with single filename-based component for testing
    processor._component_registry._cached_components = {component_id: component}

    # Construct hypothetical operation for component
    operation_name = "Filter text test"
    operation_params = {
        "text": "path/to/text.txt",
        "pattern": "hello"
    }
    operation = Operation(id='filter-text-id',
                          type='execution_node',
                          classifier=component_id,
                          name=operation_name,
                          parent_operation_ids=[],
                          component_params=operation_params)

    # Build a mock runtime config for use in _cc_pipeline
    mocked_runtime = Metadata(name="test-metadata",
                              display_name="test",
                              schema_name="airflow",
                              metadata=sample_metadata)

    mocked_func = mock.Mock(return_value="default", side_effect=[mocked_runtime, sample_metadata])
    monkeypatch.setattr(processor, "_get_metadata_configuration", mocked_func)

    # Construct single-operation pipeline
    pipeline = Pipeline(id='pipeline-id',
                        name='kfp_test',
                        runtime='kfp',
                        runtime_config='test',
                        source='filter_text.pipeline')
    pipeline.operations[operation.id] = operation

    # Establish path and function to construct pipeline
    pipeline_path = os.path.join(tmpdir, 'kfp_test.yaml')
    constructed_pipeline_function = lambda: processor._cc_pipeline(pipeline=pipeline, pipeline_name='test_pipeline')

    # TODO Check against both argo and tekton compilations
    # Compile pipeline and save into pipeline_path
    kfp_argo_compiler.Compiler().compile(constructed_pipeline_function, pipeline_path)

    # Read contents of pipeline YAML
    with open(pipeline_path) as f:
        pipeline_yaml = yaml.safe_load(f.read())

    # Check the pipeline file contents for correctness
    pipeline_template = pipeline_yaml['spec']['templates'][0]
    assert pipeline_template['metadata']['annotations']['pipelines.kubeflow.org/task_display_name'] == operation_name
    assert pipeline_template['inputs']['artifacts'][0]['raw']['data'] == operation_params['text']

    component_ref = pipeline_template['metadata']['annotations']['pipelines.kubeflow.org/component_ref']
    component_ref = ast.literal_eval(component_ref)
    assert relative_path in component_ref['url']



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from logging import Logger
import os
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

# TODO: Make pipeline version available more widely
# as today is only available on the pipeline editor
PIPELINE_CURRENT_VERSION = 5
PIPELINE_CURRENT_SCHEMA = 3.0


class Operation(object):
    """
    Represents a single operation in a pipeline representing a third-party component
    """

    generic_node_types = ["execute-notebook-node", "execute-python-node", "execute-r-node"]

    @classmethod
    def create_instance(cls, id: str, type: str, name: str, classifier: str,
                        parent_operation_ids: Optional[List[str]] = None,
                        component_params: Optional[Dict[str, Any]] = None) -> 'Operation':
        """Class method that creates the appropriate instance of Operation based on inputs. """

        if classifier in Operation.generic_node_types:
            return GenericOperation(id, type, name, classifier,
                                    parent_operation_ids=parent_operation_ids, component_params=component_params)
        return Operation(id, type, name, classifier,
                         parent_operation_ids=parent_operation_ids, component_params=component_params)

    def __init__(self, id: str, type: str, name: str, classifier: str,
                 parent_operation_ids: Optional[List[str]] = None,
                 component_params: Optional[Dict[str, Any]] = None):
        """
        :param id: Generated UUID, 128 bit number used as a unique identifier
                   e.g. 123e4567-e89b-12d3-a456-426614174000
        :param type: The type of node e.g. execution_node
        :param classifier: indicates the operation's class
        :param name: The name of the operation
        :param parent_operation_ids: List of parent operation 'ids' required to execute prior to this operation
        :param component_params: dictionary of parameter key:value pairs that are used in the creation of a
                                 a non-standard operation instance
        """

        # Validate that the operation has all required properties
        if not id:
            raise ValueError("Invalid pipeline operation: Missing field 'operation id'.")
        if not type:
            raise ValueError("Invalid pipeline operation: Missing field 'operation type'.")
        if not classifier:
            raise ValueError("Invalid pipeline operation: Missing field 'operation classifier'.")
        if not name:
            raise ValueError("Invalid pipeline operation: Missing field 'operation name'.")

        self._id = id
        self._type = type
        self._classifier = classifier
        self._name = name
        self._parent_operation_ids = parent_operation_ids or []
        self._component_params = component_params

        # Scrub the inputs and outputs lists
        self._component_params["inputs"] = Operation._scrub_list(component_params.get('inputs', []))
        self._component_params["outputs"] = Operation._scrub_list(component_params.get('outputs', []))

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> str:
        return self._type

    @property
    def classifier(self) -> str:
        return self._classifier

    @property
    def name(self) -> str:
        return self._name

    @property
    def parent_operation_ids(self) -> List[str]:
        return self._parent_operation_ids

    @property
    def component_params(self) -> Optional[Dict[str, Any]]:
        return self._component_params

    @property
    def component_params_as_dict(self) -> Dict[str, Any]:
        return self._component_params or {}

    @property
    def inputs(self) -> Optional[List[str]]:
        return self._component_params.get('inputs')

    @inputs.setter
    def inputs(self, value: List[str]):
        self._component_params['inputs'] = value

    @property
    def outputs(self) -> Optional[List[str]]:
        return self._component_params.get('outputs')

    @outputs.setter
    def outputs(self, value: List[str]):
        self._component_params['outputs'] = value

    def __eq__(self, other: 'Operation') -> bool:
        if isinstance(self, other.__class__):
            return self.id == other.id and \
                self.type == other.type and \
                self.classifier == other.classifier and \
                self.name == other.name and \
                self.parent_operation_ids == other.parent_operation_ids and \
                self.component_params == other.component_params
        return False

    def __str__(self) -> str:
        params = ""
        for key, value in self.component_params_as_dict.items():
            params += f"\t{key}: {value}, \n"

        return f"componentID : {self.id} \n " \
            f"name : {self.name} \n " \
            f"parent_operation_ids : {self.parent_operation_ids} \n " \
            f"component_parameters: {{\n{params}}} \n"

    @staticmethod
    def _log_info(msg: str, logger: Optional[Logger] = None):
        if logger:
            logger.info(msg)
        else:
            print(msg)

    @staticmethod
    def _log_warning(msg: str, logger: Optional[Logger] = None):
        if logger:
            logger.warning(msg)
        else:
            print(f"WARNING: {msg}")

    @staticmethod
    def _scrub_list(dirty: Optional[List[Optional[str]]]) -> List[str]:
        """
        Clean an existing list by filtering out None and empty string values
        :param dirty: a List of values
        :return: a clean list without None or empty string values
        """
        if not dirty:
            return []
        return [clean for clean in dirty if clean]

    @staticmethod
    def is_generic_operation(operation_type) -> bool:
        return operation_type in Operation.generic_node_types


class GenericOperation(Operation):
    """
    Represents a single operation in a pipeline representing a generic (built-in) component
    """

    def __init__(self, id: str, type: str, name: str, classifier: str,
                 parent_operation_ids: Optional[List[str]] = None,
                 component_params: Optional[Dict[str, Any]] = None):
        """
        :param id: Generated UUID, 128 bit number used as a unique identifier
                   e.g. 123e4567-e89b-12d3-a456-426614174000
        :param type: The type of node e.g. execution_node
        :param classifier: indicates the operation's class
        :param name: The name of the operation
        :param parent_operation_ids: List of parent operation 'ids' required to execute prior to this operation
        :param component_params: dictionary of parameter key:value pairs that are used in the creation of a
                                 a non-standard operation instance

        Component_params for "generic components" (i.e., those with one of the following classifier values:
        ["execute-notebook-node", "execute-python-node", "exeucute-r-node"]) can expect to have the following
        entries.
                filename: The relative path to the source file in the users local environment
                         to be executed e.g. path/to/file.ext
                runtime_image: The DockerHub image to be used for the operation
                               e.g. user/docker_image_name:tag
                dependencies: List of local files/directories needed for the operation to run
                             and packaged into each operation's dependency archive
                include_subdirectories: Include or Exclude subdirectories when packaging our 'dependencies'
                env_vars: List of Environmental variables to set in the docker image
                         e.g. FOO="BAR"
                inputs: List of files to be consumed by this operation, produced by parent operation(s)
                outputs: List of files produced by this operation to be included in a child operation(s)
                cpu: number of cpus requested to run the operation
                memory: amount of memory requested to run the operation (in Gi)
                gpu: number of gpus requested to run the operation
        Entries for other (non-built-in) component types are a function of the respective component.
        """

        super().__init__(id, type, name, classifier,
                         parent_operation_ids=parent_operation_ids, component_params=component_params)

        if not component_params.get('filename'):
            raise ValueError("Invalid pipeline operation: Missing field 'operation filename'.")
        if not component_params.get('runtime_image'):
            raise ValueError("Invalid pipeline operation: Missing field 'operation runtime image'.")
        if component_params.get('cpu') and not self._validate_range(component_params.get('cpu'), min_value=1):
            raise ValueError("Invalid pipeline operation: CPU must be a positive value or None")
        if component_params.get('gpu') and not self._validate_range(component_params.get('gpu'), min_value=0):
            raise ValueError("Invalid pipeline operation: GPU must be a positive value or None")
        if component_params.get('memory') and not self._validate_range(component_params.get('memory'), min_value=1):
            raise ValueError("Invalid pipeline operation: Memory must be a positive value or None")

        # Re-build object to include default values
        self._component_params["filename"] = component_params.get('filename')
        self._component_params["runtime_image"] = component_params.get('runtime_image')
        self._component_params["dependencies"] = Operation._scrub_list(component_params.get('dependencies', []))
        self._component_params["include_subdirectories"] = component_params.get('include_subdirectories', False)
        self._component_params["env_vars"] = Operation._scrub_list(component_params.get('env_vars', []))
        self._component_params["cpu"] = component_params.get('cpu')
        self._component_params["gpu"] = component_params.get('gpu')
        self._component_params["memory"] = component_params.get('memory')

    @property
    def name(self) -> str:
        if self._name == os.path.basename(self.filename):
            self._name = os.path.basename(self._name).split(".")[0]
        return self._name

    @property
    def filename(self) -> str:
        return self._component_params.get('filename')

    @property
    def runtime_image(self) -> str:
        return self._component_params.get('runtime_image')

    @property
    def dependencies(self) -> Optional[List[str]]:
        return self._component_params.get('dependencies')

    @property
    def include_subdirectories(self) -> Optional[bool]:
        return self._component_params.get('include_subdirectories')

    @property
    def env_vars(self) -> Optional[List[str]]:
        return self._component_params.get('env_vars')

    @property
    def cpu(self) -> Optional[str]:
        return self._component_params.get('cpu')

    @property
    def memory(self) -> Optional[str]:
        return self._component_params.get('memory')

    @property
    def gpu(self) -> Optional[str]:
        return self._component_params.get('gpu')

    def __eq__(self, other: 'GenericOperation') -> bool:
        if isinstance(self, other.__class__):
            return super().__eq__(other)
        return False

    def _validate_range(self, value: str, min_value: int = 0, max_value: int = sys.maxsize) -> bool:
        return int(value) in range(min_value, max_value)

    def env_vars_as_dict(self, logger: Optional[Logger] = None) -> Dict[str, str]:
        """
        Operation stores environment variables in a list of name=value pairs, while
        subprocess.run() requires a dictionary - so we must convert.  If no envs are
        configured on the Operation, an empty dictionary is returned, otherwise envs
        configured on the Operation are converted to dictionary entries and returned.
        """
        envs = {}
        for nv in self.env_vars:
            if nv:
                nv_pair = nv.split("=", 1)
                if len(nv_pair) == 2 and nv_pair[0].strip():
                    if len(nv_pair[1]) > 0:
                        envs[nv_pair[0]] = nv_pair[1]
                    else:
                        Operation._log_info(f"Skipping inclusion of environment variable: "
                                            f"`{nv_pair[0]}` has no value...",
                                            logger=logger)
                else:
                    Operation._log_warning(f"Could not process environment variable entry `{nv}`, skipping...",
                                           logger=logger)
        return envs


class Pipeline(object):
    """
    Represents a single pipeline constructed in the pipeline editor
    """

    def __init__(self,
                 id: str,
                 name: str,
                 runtime: str,
                 runtime_config: str,
                 source: Optional[str] = None,
                 description: Optional[str] = None):
        """
        :param id: Generated UUID, 128 bit number used as a unique identifier
                   e.g. 123e4567-e89b-12d3-a456-426614174000
        :param name: Pipeline name
                     e.g. test-pipeline-123456
        :param runtime: Type of runtime we want to use to execute our pipeline
                        e.g. kfp OR airflow
        :param runtime_config: Runtime configuration that should be used to submit the pipeline to execution
        :param source: The pipeline source, e.g. a pipeline file or a notebook.
        :description: Pipeline description
        """

        if not name:
            raise ValueError('Invalid pipeline: Missing pipeline name.')
        if not runtime:
            raise ValueError('Invalid pipeline: Missing runtime.')
        if not runtime_config:
            raise ValueError('Invalid pipeline: Missing runtime configuration.')

        self._id = id
        self._name = name
        self._description = description
        self._source = source
        self._runtime = runtime
        self._runtime_config = runtime_config
        self._operations = {}

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def source(self) -> str:
        return self._source

    @property
    def runtime(self) -> str:
        """
        Describe the runtime type where the pipeline will be executed
        """
        return self._runtime

    @property
    def runtime_config(self) -> str:
        """
        Describe the runtime configuration that should be used to submit the pipeline to execution
        """
        return self._runtime_config

    @property
    def operations(self) -> Dict[str, Operation]:
        return self._operations

    @property
    def description(self) -> Optional[str]:
        """
        Pipeline description
        """
        return self._description

    def __eq__(self, other: 'Pipeline') -> bool:
        if isinstance(self, other.__class__):
            return self.id == other.id and \
                self.name == other.name and \
                self.source == other.source and \
                self.description == other.description and \
                self.runtime == other.runtime and \
                self.runtime_config == other.runtime_config and \
                self.operations == other.operations



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC
from abc import abstractmethod
import glob
import json
import logging
import os
from pathlib import Path
import subprocess
import sys
from tempfile import TemporaryFile
import time
from typing import Any
from typing import Optional
from typing import Type
from typing import TypeVar
from urllib.parse import urljoin
from urllib.parse import urlparse
from urllib.parse import urlunparse

from packaging import version


# Inputs and Outputs separator character.  If updated,
# same-named variable in _notebook_op.py must be updated!
INOUT_SEPARATOR = ';'

# Setup forward reference for type hint on return from class factory method.  See
# https://stackoverflow.com/questions/39205527/can-you-annotate-return-type-when-value-is-instance-of-cls/39205612#39205612
F = TypeVar('F', bound='FileOpBase')

logger = logging.getLogger('elyra')
enable_pipeline_info = os.getenv('ELYRA_ENABLE_PIPELINE_INFO', 'true').lower() == 'true'
pipeline_name = None  # global used in formatted logging
operation_name = None  # global used in formatted logging


class FileOpBase(ABC):
    """Abstract base class for file-based operations"""
    filepath = None
    cos_client = None
    cos_bucket = None

    @classmethod
    def get_instance(cls: Type[F], **kwargs: Any) -> F:
        """Creates an appropriate subclass instance based on the extension of the filepath (-f) argument"""
        filepath = kwargs['filepath']
        if '.ipynb' in filepath:
            return NotebookFileOp(**kwargs)
        elif '.py' in filepath:
            return PythonFileOp(**kwargs)
        elif '.r' in filepath:
            return RFileOp(**kwargs)
        else:
            raise ValueError('Unsupported file type: {}'.format(filepath))

    def __init__(self, **kwargs: Any) -> None:
        """Initializes the FileOpBase instance"""
        import minio

        self.filepath = kwargs['filepath']
        self.input_params = kwargs or []
        self.cos_endpoint = urlparse(self.input_params.get('cos-endpoint'))
        self.cos_bucket = self.input_params.get('cos-bucket')

        # Infer secure from the endpoint's scheme.
        self.secure = self.cos_endpoint.scheme == 'https'

        self.cos_client = minio.Minio(self.cos_endpoint.netloc,
                                      access_key=os.getenv('AWS_ACCESS_KEY_ID'),
                                      secret_key=os.getenv('AWS_SECRET_ACCESS_KEY'),
                                      secure=self.secure)

    @abstractmethod
    def execute(self) -> None:
        """Execute the operation relative to derived class"""
        raise NotImplementedError("Method 'execute()' must be implemented by subclasses!")

    def process_dependencies(self) -> None:
        """Process dependencies

        If a dependency archive is present, it will be downloaded from object storage
        and expanded into the local directory.

        This method can be overridden by subclasses, although overrides should first
        call the superclass method.
        """
        OpUtil.log_operation_info('processing dependencies')
        t0 = time.time()
        archive_file = self.input_params.get('cos-dependencies-archive')

        self.get_file_from_object_storage(archive_file)

        inputs = self.input_params.get('inputs')
        if inputs:
            input_list = inputs.split(INOUT_SEPARATOR)
            for file in input_list:
                self.get_file_from_object_storage(file.strip())

        subprocess.call(['tar', '-zxvf', archive_file])
        duration = time.time() - t0
        OpUtil.log_operation_info("dependencies processed", duration)

    def process_outputs(self) -> None:
        """Process outputs

        If outputs have been specified, it will upload the appropriate files to object storage

        This method can be overridden by subclasses, although overrides should first
        call the superclass method.
        """
        OpUtil.log_operation_info('processing outputs')
        t0 = time.time()
        outputs = self.input_params.get('outputs')
        if outputs:
            output_list = outputs.split(INOUT_SEPARATOR)
            for file in output_list:
                self.process_output_file(file.strip())
        duration = time.time() - t0
        OpUtil.log_operation_info('outputs processed', duration)

    def process_metrics_and_metadata(self) -> None:
        """Process metrics and metadata

        This method exposes metrics/metadata that the processed
        notebook | script produces in the KFP UI.

        This method should not be overridden by subclasses.
        """

        OpUtil.log_operation_info('processing metrics and metadata')
        t0 = time.time()

        # Location where the KFP specific output files will be stored
        # in the environment where the bootsrapper is running.
        # Defaults to '/tmp' if not specified.
        output_path = Path(os.getenv('ELYRA_WRITABLE_CONTAINER_DIR', '/tmp'))

        # verify that output_path exists, is a directory
        # and writable by creating a temporary file in that location
        try:
            with TemporaryFile(mode='w', dir=output_path) as t:
                t.write('can write')
        except Exception:
            # output_path doesn't meet the requirements
            # treat this as a non-fatal error and log a warning
            logger.warning('Cannot create files in "{}".'
                           .format(output_path))
            OpUtil.log_operation_info('Aborted metrics and metadata processing',
                                      time.time() - t0)
            return

        # Name of the proprietary KFP UI metadata file.
        # Notebooks | scripts might (but don't have to) produce this file
        # as documented in
        # https://www.kubeflow.org/docs/pipelines/sdk/output-viewer/
        # Each ExecuteFileOp must declare this as an output file or
        # the KFP UI won't pick up the information.
        kfp_ui_metadata_filename = 'mlpipeline-ui-metadata.json'

        # Name of the proprietary KFP metadata file.
        # Notebooks | scripts might (but don't have to) produce this file
        # as documented in
        # https://www.kubeflow.org/docs/pipelines/sdk/pipelines-metrics/
        # Each ExecuteFileOp must declare this as an output file or
        # the KFP UI won't pick up the information.
        kfp_metrics_filename = 'mlpipeline-metrics.json'

        # If the notebook | Python script produced one of the files
        # copy it to the target location where KFP is looking for it.
        for filename in [kfp_ui_metadata_filename, kfp_metrics_filename]:
            try:
                src = Path('.') / filename
                logger.debug('Processing {} ...'.format(src))
                # try to load the file, if one was created by the
                # notebook or script
                with open(src, 'r') as f:
                    metadata = json.load(f)

                # the file exists and contains valid JSON
                logger.debug('File content: {}'.format(json.dumps(metadata)))

                target = output_path / filename
                # try to save the file in the destination location
                with open(target, 'w') as f:
                    json.dump(metadata, f)
            except FileNotFoundError:
                # The script | notebook didn't produce the file
                # we are looking for. This is not an error condition
                # that needs to be handled.
                logger.debug('{} produced no file named {}'
                             .format(self.filepath,
                                     src))
            except ValueError as ve:
                # The file content could not be parsed. Log a warning
                # and treat this as a non-fatal error.
                logger.warning('Ignoring incompatible {} produced by {}: {} {}'.
                               format(str(src),
                                      self.filepath,
                                      ve,
                                      str(ve)))
            except Exception as ex:
                # Something is wrong with the user-generated metadata file.
                # Log a warning and treat this as a non-fatal error.
                logger.warning('Error processing {} produced by {}: {} {}'.
                               format(str(src),
                                      self.filepath,
                                      ex,
                                      str(ex)))

        #
        # Augment kfp_ui_metadata_filename with Elyra-specific information:
        #  - link to object storage where input and output artifacts are
        #    stored
        ui_metadata_output = output_path / kfp_ui_metadata_filename
        try:
            # re-load the file
            with open(ui_metadata_output, 'r') as f:
                metadata = json.load(f)
        except Exception:
            # ignore all errors
            metadata = {}

        # Assure the 'output' property exists and is of the correct type
        if metadata.get('outputs', None) is None or\
           not isinstance(metadata['outputs'], list):
            metadata['outputs'] = []

        # Define HREF for COS bucket:
        # <COS_URL>/<BUCKET_NAME>/<COS_DIRECTORY>
        bucket_url =\
            urljoin(urlunparse(self.cos_endpoint),
                    '{}/{}/'
                    .format(self.cos_bucket,
                            self.input_params.get('cos-directory', '')))

        # add Elyra metadata to 'outputs'
        metadata['outputs'].append({
            'storage': 'inline',
            'source': '## Inputs for {}\n'
                      '[{}]({})'
                      .format(self.filepath,
                              self.input_params['cos-dependencies-archive'],
                              bucket_url),
            'type': 'markdown'
        })

        # print the content of the augmented metadata file
        logger.debug('Output UI metadata: {}'.format(json.dumps(metadata)))

        logger.debug('Saving UI metadata file as {} ...'
                     .format(ui_metadata_output))

        # Save [updated] KFP UI metadata file
        with open(ui_metadata_output, 'w') as f:
            json.dump(metadata, f)

        duration = time.time() - t0
        OpUtil.log_operation_info('metrics and metadata processed', duration)

    def get_object_storage_filename(self, filename: str) -> str:
        """Function to pre-pend cloud storage working dir to file name

        :param filename: the local file
        :return: the full path of the object storage file
        """
        return os.path.join(self.input_params.get('cos-directory', ''), filename)

    def get_file_from_object_storage(self, file_to_get: str) -> None:
        """Utility function to get files from an object storage

        :param file_to_get: filename
        """

        object_to_get = self.get_object_storage_filename(file_to_get)
        t0 = time.time()
        self.cos_client.fget_object(bucket_name=self.cos_bucket,
                                    object_name=object_to_get,
                                    file_path=file_to_get)
        duration = time.time() - t0
        OpUtil.log_operation_info(f"downloaded {file_to_get} from bucket: {self.cos_bucket}, object: {object_to_get}",
                                  duration)

    def put_file_to_object_storage(self, file_to_upload: str, object_name: Optional[str] = None) -> None:
        """Utility function to put files into an object storage

        :param file_to_upload: filename
        :param object_name: remote filename (used to rename)
        """

        object_to_upload = object_name
        if not object_to_upload:
            object_to_upload = file_to_upload

        object_to_upload = self.get_object_storage_filename(object_to_upload)
        t0 = time.time()
        self.cos_client.fput_object(bucket_name=self.cos_bucket,
                                    object_name=object_to_upload,
                                    file_path=file_to_upload)
        duration = time.time() - t0
        OpUtil.log_operation_info(f"uploaded {file_to_upload} to bucket: {self.cos_bucket} object: {object_to_upload}",
                                  duration)

    def has_wildcard(self, filename):
        wildcards = ['*', '?']
        return bool(any(c in filename for c in wildcards))

    def process_output_file(self, output_file):
        """Puts the file to object storage.  Handles wildcards and directories. """

        matched_files = [output_file]
        if self.has_wildcard(output_file):  # explode the wildcarded file
            matched_files = glob.glob(output_file)

        for matched_file in matched_files:
            if os.path.isdir(matched_file):
                for file in os.listdir(matched_file):
                    self.process_output_file(os.path.join(matched_file, file))
            else:
                self.put_file_to_object_storage(matched_file)


class NotebookFileOp(FileOpBase):
    """Perform Notebook File Operation"""

    def execute(self) -> None:
        """Execute the Notebook and upload results to object storage"""
        notebook = os.path.basename(self.filepath)
        notebook_name = notebook.replace('.ipynb', '')
        notebook_output = notebook_name + '-output.ipynb'
        notebook_html = notebook_name + '.html'

        try:
            OpUtil.log_operation_info(f"executing notebook using 'papermill {notebook} {notebook_output}'")
            t0 = time.time()
            # Include kernel selection in execution time
            kernel_name = NotebookFileOp.find_best_kernel(notebook)

            import papermill
            papermill.execute_notebook(notebook, notebook_output, kernel_name=kernel_name)
            duration = time.time() - t0
            OpUtil.log_operation_info("notebook execution completed", duration)

            NotebookFileOp.convert_notebook_to_html(notebook_output, notebook_html)
            self.put_file_to_object_storage(notebook_output, notebook)
            self.put_file_to_object_storage(notebook_html)
            self.process_outputs()
        except Exception as ex:
            # log in case of errors
            logger.error("Unexpected error: {}".format(sys.exc_info()[0]))

            NotebookFileOp.convert_notebook_to_html(notebook_output, notebook_html)
            self.put_file_to_object_storage(notebook_output, notebook)
            self.put_file_to_object_storage(notebook_html)
            raise ex

    @staticmethod
    def convert_notebook_to_html(notebook_file: str, html_file: str) -> str:
        """Function to convert a Jupyter notebook file (.ipynb) into an html file

        :param notebook_file: object storage client
        :param html_file: name of what the html output file should be
        :return: html_file: the converted notebook in html format
        """
        import nbconvert
        import nbformat

        OpUtil.log_operation_info(f"converting from {notebook_file} to {html_file}")
        t0 = time.time()
        nb = nbformat.read(notebook_file, as_version=4)
        html_exporter = nbconvert.HTMLExporter()
        data, resources = html_exporter.from_notebook_node(nb)
        with open(html_file, "w") as f:
            f.write(data)
            f.close()

        duration = time.time() - t0
        OpUtil.log_operation_info(f"{notebook_file} converted to {html_file}", duration)
        return html_file

    @staticmethod
    def find_best_kernel(notebook_file: str) -> str:
        """Determines the best kernel to use via the following algorithm:

           1. Loads notebook and gets kernel_name and kernel_language from NB metadata.
           2. Gets the list of configured kernels using KernelSpecManager.
           3. If notebook kernel_name is in list, use that, else
           4. If not found, load each configured kernel.json file and find a language match.
           5. On first match, log info message regarding the switch and use that kernel.
           6. If no language match is found, revert to notebook kernel and log warning message.
        """
        import json
        from jupyter_client.kernelspec import KernelSpecManager
        import nbformat

        nb = nbformat.read(notebook_file, 4)

        nb_kspec = nb.metadata.kernelspec
        nb_kernel_name = nb_kspec.get('name')
        nb_kernel_lang = nb_kspec.get('language')

        kernel_specs = KernelSpecManager().find_kernel_specs()

        # see if we have a direct match...
        if nb_kernel_name in kernel_specs.keys():
            return nb_kernel_name

        # no match found for kernel, try matching language...
        for name, file in kernel_specs.items():
            # load file (JSON) and pick out language, if match, use first found
            with open(os.path.join(file, 'kernel.json')) as f:
                kspec = json.load(f)
                if kspec.get('language').lower() == nb_kernel_lang.lower():
                    matched_kernel = os.path.basename(file)
                    logger.info(f"Matched kernel by language ({nb_kernel_lang}), using kernel "
                                f"'{matched_kernel}' instead of the missing kernel '{nb_kernel_name}'.")
                    return matched_kernel

        # no match found for language, return notebook kernel and let execution fail
        logger.warning(f"Reverting back to missing notebook kernel '{nb_kernel_name}' since no "
                       f"language match ({nb_kernel_lang}) was found in current kernel specifications.")
        return nb_kernel_name


class PythonFileOp(FileOpBase):
    """Perform Python File Operation"""

    def execute(self) -> None:
        """Execute the Python script and upload results to object storage"""
        python_script = os.path.basename(self.filepath)
        python_script_name = python_script.replace('.py', '')
        python_script_output = python_script_name + '.log'

        try:
            OpUtil.log_operation_info(f"executing python script using "
                                      f"'python3 {python_script}' to '{python_script_output}'")
            t0 = time.time()
            with open(python_script_output, "w") as log_file:
                subprocess.run(['python3', python_script], stdout=log_file, stderr=subprocess.STDOUT, check=True)

            duration = time.time() - t0
            OpUtil.log_operation_info("python script execution completed", duration)

            self.put_file_to_object_storage(python_script_output, python_script_output)
            self.process_outputs()
        except Exception as ex:
            # log in case of errors
            logger.error("Unexpected error: {}".format(sys.exc_info()[0]))
            logger.error("Error details: {}".format(ex))

            self.put_file_to_object_storage(python_script_output, python_script_output)
            raise ex


class RFileOp(FileOpBase):
    """Perform R File Operation"""

    def execute(self) -> None:
        """Execute the R script and upload results to object storage"""
        r_script = os.path.basename(self.filepath)
        r_script_name = r_script.replace('.r', '')
        r_script_output = r_script_name + '.log'

        try:
            OpUtil.log_operation_info(f"executing R script using "
                                      f"'Rscript {r_script}' to '{r_script_output}'")
            t0 = time.time()
            with open(r_script_output, "w") as log_file:
                subprocess.run(['Rscript', r_script], stdout=log_file, stderr=subprocess.STDOUT, check=True)

            duration = time.time() - t0
            OpUtil.log_operation_info("R script execution completed", duration)

            self.put_file_to_object_storage(r_script_output, r_script_output)
            self.process_outputs()
        except Exception as ex:
            # log in case of errors
            logger.error("Unexpected error: {}".format(sys.exc_info()[0]))
            logger.error("Error details: {}".format(ex))

            self.put_file_to_object_storage(r_script_output, r_script_output)
            raise ex


class OpUtil(object):
    """Utility functions for preparing file execution."""
    @classmethod
    def package_install(cls, user_volume_path) -> None:
        OpUtil.log_operation_info("Installing packages")
        t0 = time.time()
        elyra_packages = cls.package_list_to_dict("requirements-elyra.txt")
        current_packages = cls.package_list_to_dict("requirements-current.txt")
        to_install_list = []

        for package, ver in elyra_packages.items():
            if package in current_packages:
                if "git+" in current_packages[package]:
                    logger.warning(f"WARNING: Source package {package} found already installed from "
                                   f"{current_packages[package]}. This may conflict with the required "
                                   f"version: {ver} . Skipping...")
                elif isinstance(version.parse(current_packages[package]), version.LegacyVersion):
                    logger.warning(f"WARNING: Package {package} found with unsupported Legacy version "
                                   f"scheme {current_packages[package]} already installed. Skipping...")
                elif version.parse(ver) > version.parse(current_packages[package]):
                    logger.info(f"Updating {package} package from version {current_packages[package]} to {ver}...")
                    to_install_list.append(package + '==' + ver)
                elif version.parse(ver) < version.parse(current_packages[package]):
                    logger.info(f"Newer {package} package with version {current_packages[package]} "
                                f"already installed. Skipping...")
            else:
                logger.info(f"Package not found. Installing {package} package with version {ver}...")
                to_install_list.append(package + '==' + ver)

        if to_install_list:
            if user_volume_path:
                to_install_list.insert(0, '--target=' + user_volume_path)
                to_install_list.append('--no-cache-dir')

            subprocess.run([sys.executable, '-m', 'pip', 'install'] + to_install_list, check=True)

        if user_volume_path:
            os.environ["PIP_CONFIG_FILE"] = user_volume_path + "/pip.conf"

        subprocess.run([sys.executable, '-m', 'pip', 'freeze'])
        duration = time.time() - t0
        OpUtil.log_operation_info("Packages installed", duration)

    @classmethod
    def package_list_to_dict(cls, filename: str) -> dict:
        package_dict = {}
        with open(filename) as fh:
            for line in fh:
                if line[0] != '#':
                    if " @ " in line:
                        package_name, package_version = line.strip('\n').split(sep=" @ ")
                    elif "===" in line:
                        package_name, package_version = line.strip('\n').split(sep="===")
                    else:
                        package_name, package_version = line.strip('\n').split(sep="==")

                    package_dict[package_name] = package_version

        return package_dict

    @classmethod
    def parse_arguments(cls, args) -> dict:
        import argparse
        global pipeline_name, operation_name

        logger.debug("Parsing Arguments.....")
        parser = argparse.ArgumentParser()
        parser.add_argument('-e', '--cos-endpoint', dest="cos-endpoint", help='Cloud object storage endpoint',
                            required=True)
        parser.add_argument('-b', '--cos-bucket', dest="cos-bucket", help='Cloud object storage bucket to use',
                            required=True)
        parser.add_argument('-d', '--cos-directory', dest="cos-directory",
                            help='Working directory in cloud object storage bucket to use', required=True)
        parser.add_argument('-t', '--cos-dependencies-archive', dest="cos-dependencies-archive",
                            help='Archive containing notebook and dependency artifacts', required=True)
        parser.add_argument('-f', '--file', dest="filepath", help='File to execute', required=True)
        parser.add_argument('-o', '--outputs', dest="outputs", help='Files to output to object store', required=False)
        parser.add_argument('-i', '--inputs', dest="inputs", help='Files to pull in from parent node', required=False)
        parser.add_argument('-p', '--user-volume-path', dest="user-volume-path",
                            help='Directory in Volume to install python libraries into', required=False)
        parsed_args = vars(parser.parse_args(args))

        # cos-directory is the pipeline name, set as global
        pipeline_name = parsed_args.get('cos-directory')
        # operation/node name is the basename of the non-suffixed filepath, set as global
        operation_name = os.path.basename(os.path.splitext(parsed_args.get('filepath'))[0])

        return parsed_args

    @classmethod
    def log_operation_info(cls, action_clause: str, duration_secs: Optional[float] = None) -> None:
        """Produces a formatted log INFO message used entirely for support purposes.

        This method is intended to be called for any entries that should be captured across aggregated
        log files to identify steps within a given pipeline and each of its operations.  As a result,
        calls to this method should produce single-line entries in the log (no embedded newlines).
        Each entry is prefixed with the pipeline name.

        General logging should NOT use this method but use logger.<level>() statements directly.

        :param action_clause: str representing the action that is being logged
        :param duration_secs: optional float value representing the duration of the action being logged
        """
        global pipeline_name, operation_name
        if enable_pipeline_info:
            duration_clause = f"({duration_secs:.3f} secs)" if duration_secs else ""
            logger.info(f"'{pipeline_name}':'{operation_name}' - {action_clause} {duration_clause}")


def main():
    # Configure logger format, level
    logging.basicConfig(format='[%(levelname)1.1s %(asctime)s.%(msecs).03d] %(message)s',
                        datefmt='%H:%M:%S',
                        level=logging.DEBUG)
    # Setup packages and gather arguments
    input_params = OpUtil.parse_arguments(sys.argv[1:])
    OpUtil.log_operation_info("starting operation")
    t0 = time.time()
    OpUtil.package_install(user_volume_path=input_params.get('user-volume-path'))

    # Create the appropriate instance, process dependencies and execute the operation
    file_op = FileOpBase.get_instance(**input_params)

    file_op.process_dependencies()

    file_op.execute()

    # Process notebook | script metrics and KFP UI metadata
    file_op.process_metrics_and_metadata()

    duration = time.time() - t0
    OpUtil.log_operation_info("operation completed", duration)


if __name__ == '__main__':
    main()



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for elyra-metadata application"""
import json
import os
import shutil
from tempfile import mkdtemp

import pytest

from elyra.metadata.manager import MetadataManager
from elyra.metadata.metadata import Metadata
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE_ID
from elyra.tests.metadata.test_utils import another_metadata_json
from elyra.tests.metadata.test_utils import create_json_file
from elyra.tests.metadata.test_utils import invalid_metadata_json
from elyra.tests.metadata.test_utils import invalid_no_display_name_json
from elyra.tests.metadata.test_utils import invalid_schema_name_json
from elyra.tests.metadata.test_utils import PropertyTester
from elyra.tests.metadata.test_utils import valid_metadata_json

os.environ["METADATA_TESTING"] = "1"  # Enable metadata-tests schemaspace


@pytest.fixture()
def mock_data_dir():
    runtime_dir = mkdtemp(prefix="runtime_")
    orig_data_dir = os.environ.get("JUPYTER_DATA_DIR")
    os.environ["JUPYTER_DATA_DIR"] = runtime_dir
    yield runtime_dir  # provide the fixture value
    shutil.rmtree(runtime_dir)
    if orig_data_dir:
        os.environ["JUPYTER_DATA_DIR"] = orig_data_dir
    else:
        os.environ.pop("JUPYTER_DATA_DIR")


def test_no_opts(script_runner):
    ret = script_runner.run('elyra-metadata')
    assert ret.success is False
    assert "No subcommand specified. Must specify one of: ['list', 'install', 'remove']" in ret.stdout


def test_bad_subcommand(script_runner):
    ret = script_runner.run('elyra-metadata', 'bogus-subcommand')
    assert ret.success is False
    assert ret.stdout.startswith("Subcommand 'bogus-subcommand' is invalid.")
    assert "No subcommand specified. Must specify one of: ['list', 'install', 'remove']" in ret.stdout


def test_install_bad_argument(script_runner):
    ret = script_runner.run('elyra-metadata', 'install', '--bogus-argument')
    assert ret.success is False
    assert ret.stdout.startswith("Subcommand '--bogus-argument' is invalid.")
    assert "Install a metadata instance into schemaspace \'{}\'.".format(METADATA_TEST_SCHEMASPACE) in ret.stdout


def test_install_bad_schemaspace(script_runner):
    ret = script_runner.run('elyra-metadata', 'install', 'bogus-schemaspace')
    assert ret.success is False
    assert ret.stdout.startswith("Subcommand 'bogus-schemaspace' is invalid.")
    assert "Install a metadata instance into a given schemaspace." in ret.stdout
    assert "Install a metadata instance into schemaspace \'{}\'.".format(METADATA_TEST_SCHEMASPACE) in ret.stdout


def test_install_help(script_runner):
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--help')
    assert ret.success is False
    assert ret.stdout.startswith(f"\nInstall a metadata instance into schemaspace '{METADATA_TEST_SCHEMASPACE}'.")


def test_install_no_schema_single(script_runner, mock_data_dir):
    # Use the runtime-images schemaspace since that is most likely to always be a single-schema schemaspace.
    # Note: this test will break if it ever supports multiple.
    ret = script_runner.run('elyra-metadata', 'install', "runtime-images")
    assert ret.success is False
    assert ret.stdout.startswith("'--display_name' is a required parameter.")


def test_install_no_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    # Since order in dictionaries, where the one-of list is derived, can be random, just check up to the
    # first known difference in the schema names.
    assert ret.stdout.startswith("'--schema_name' is a required parameter and must be one of the "
                                 "following values: ['metadata-test")


def test_install_bad_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-foo')
    assert ret.success is False
    assert ret.stdout.startswith("Parameter '--schema_name' requires one of the following values: ['metadata-test")


def test_install_no_name(script_runner, mock_data_dir):
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test')
    assert ret.success is False
    assert ret.stdout.startswith("'--display_name' is a required parameter.")


def test_install_only_display_name(script_runner, mock_data_dir):
    metadata_display_name = "1 teste 'rápido'"
    metadata_name = 'a_1_teste_rpido'

    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            f'--display_name={metadata_display_name}', '--required_test=required_value')
    assert ret.success is True
    assert ret.stdout.startswith(f"Metadata instance '{metadata_name}' for schema 'metadata-test' has been written to:")

    # Ensure it can be fetched by name...
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE_ID)
    resource = metadata_manager.get(metadata_name)
    assert resource.display_name == metadata_display_name


def test_install_invalid_name(script_runner, mock_data_dir):
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=UPPER_CASE_NOT_ALLOWED', '--display_name=display_name',
                            '--required_test=required_value')
    assert ret.success is False
    assert ret.stdout.startswith("The following exception occurred saving metadata instance for schema 'metadata-test'")
    assert "Name of metadata must be lowercase alphanumeric" in ret.stdout


def test_install_simple(script_runner, mock_data_dir):
    expected_file = os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE,
                                 'test-metadata_42_valid-name.json')
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=test-metadata_42_valid-name', '--display_name=display_name',
                            '--required_test=required_value')
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == 'metadata-test'
        assert instance_json["display_name"] == 'display_name'
        assert instance_json["metadata"]["required_test"] == 'required_value'
        assert instance_json["metadata"]["number_default_test"] == 42  # defaults will always persist


def test_install_and_replace(script_runner, mock_data_dir):
    expected_file = os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE,
                                 'test-metadata_42_valid-name.json')
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=test-metadata_42_valid-name', '--display_name=display_name',
                            '--required_test=required_value')
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout

    # Re-attempt w/o replace flag - failure expected
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=test-metadata_42_valid-name', '--display_name=display_name',
                            '--required_test=required_value')
    assert ret.success is False
    assert "An instance named 'test-metadata_42_valid-name' already exists in the metadata-tests " \
           "schemaspace" in ret.stderr

    # Re-attempt with replace flag but without --name - failure expected
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--display_name=display_name', '--required_test=required_value', '--replace')
    assert ret.success is False
    assert "Name of metadata was not provided" in ret.stdout

    # Re-attempt with replace flag - success expected
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=test-metadata_42_valid-name', '--display_name=display_name',
                            '--required_test=required_value', '--replace')
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == 'metadata-test'
        assert instance_json["display_name"] == 'display_name'
        assert instance_json["metadata"]["required_test"] == 'required_value'
        assert instance_json["metadata"]["number_default_test"] == 42  # defaults will always persist


def test_list_help(script_runner):
    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE, '--help')
    assert ret.success is False
    assert ret.stdout.startswith("\nList installed metadata for {}.".format(METADATA_TEST_SCHEMASPACE))


def test_list_bad_argument(script_runner):
    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE, '--bogus-argument')
    assert ret.success is False
    assert ret.stdout.startswith("The following arguments were unexpected: ['--bogus-argument']")


def test_list_instances(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE)
    assert ret.success
    lines = ret.stdout.split('\n')
    assert len(lines) == 2  # always 2 more than the actual runtime count
    assert lines[0].startswith("No metadata instances found for {}".format(METADATA_TEST_SCHEMASPACE))

    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create('valid', valid)
    assert resource is not None
    resource = metadata_manager.create('valid2', valid)
    assert resource is not None
    another = Metadata(**another_metadata_json)
    resource = metadata_manager.create('another', another)
    assert resource is not None
    resource = metadata_manager.create('another2', another)
    assert resource is not None

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE)
    assert ret.success
    lines = ret.stdout.split('\n')
    assert len(lines) == 9  # always 5 more than the actual runtime count
    assert lines[0] == "Available metadata instances for {} (includes invalid):".format(METADATA_TEST_SCHEMASPACE)
    line_elements = [line.split() for line in lines[4:8]]
    assert line_elements[0][0] == "metadata-test"
    assert line_elements[0][1] == "another"
    assert line_elements[1][0] == "metadata-test"
    assert line_elements[1][1] == "another2"
    assert line_elements[2][0] == "metadata-test"
    assert line_elements[2][1] == "valid"
    assert line_elements[3][0] == "metadata-test"
    assert line_elements[3][1] == "valid2"

    # Remove the '2' runtimes and reconfirm smaller set
    metadata_manager.remove('valid2')
    metadata_manager.remove('another2')
    # Include two additional invalid files as well - one for uri failure, andother missing display_name
    metadata_dir = os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE)
    create_json_file(metadata_dir, 'invalid.json', invalid_metadata_json)
    create_json_file(metadata_dir, 'no_display_name.json', invalid_no_display_name_json)
    create_json_file(metadata_dir, 'invalid_schema_name.json', invalid_schema_name_json)

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE)
    assert ret.success
    lines = ret.stdout.split('\n')
    assert len(lines) == 10  # always 5 more than the actual runtime count
    assert lines[0] == "Available metadata instances for {} (includes invalid):".format(METADATA_TEST_SCHEMASPACE)
    line_elements = [line.split() for line in lines[4:9]]
    assert line_elements[0][1] == "another"
    assert line_elements[1][1] == "invalid"
    assert line_elements[1][3] == "**INVALID**"
    assert line_elements[1][4] == "(ValidationError)"
    assert line_elements[2][3] == "**INVALID**"
    assert line_elements[2][4] == "(ValidationError)"
    assert line_elements[3][1] == "valid"
    assert line_elements[4][3] == "**INVALID**"
    assert line_elements[4][4] == "(SchemaNotFoundError)"

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE, '--valid-only')
    assert ret.success
    lines = ret.stdout.split('\n')
    assert len(lines) == 7  # always 5 more than the actual runtime count
    assert lines[0] == "Available metadata instances for {} (valid only):".format(METADATA_TEST_SCHEMASPACE)
    line_elements = [line.split() for line in lines[4:6]]
    assert line_elements[0][1] == "another"
    assert line_elements[1][1] == "valid"


def test_list_json_instances(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE, '--json')
    assert ret.success
    lines = ret.stdout.split('\n')
    assert len(lines) == 2
    assert lines[0] == "[]"

    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create('valid', valid)
    assert resource is not None
    resource = metadata_manager.create('valid2', valid)
    assert resource is not None
    another = Metadata(**another_metadata_json)
    resource = metadata_manager.create('another', another)
    assert resource is not None
    resource = metadata_manager.create('another2', another)
    assert resource is not None

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE, '--json')
    assert ret.success
    # Consume results
    results = json.loads(ret.stdout)
    assert len(results) == 4

    # Remove the '2' runtimes and reconfirm smaller set
    metadata_manager.remove('valid2')
    metadata_manager.remove('another2')

    # Include two additional invalid files as well - one for uri failure, andother missing display_name
    metadata_dir = os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE)
    create_json_file(metadata_dir, 'invalid.json', invalid_metadata_json)
    create_json_file(metadata_dir, 'no_display_name.json', invalid_no_display_name_json)

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE, '--json')
    assert ret.success
    results = json.loads(ret.stdout)
    assert len(results) == 4

    ret = script_runner.run('elyra-metadata', 'list', METADATA_TEST_SCHEMASPACE, '--json', '--valid-only')
    assert ret.success
    results = json.loads(ret.stdout)
    assert len(results) == 2


def test_remove_help(script_runner):
    ret = script_runner.run('elyra-metadata', 'remove', METADATA_TEST_SCHEMASPACE, '--help')
    assert ret.success is False
    assert ret.stdout.startswith(f"\nRemove a metadata instance from schemaspace '{METADATA_TEST_SCHEMASPACE}'.")


def test_remove_no_name(script_runner):
    ret = script_runner.run('elyra-metadata', 'remove', METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    assert ret.stdout.startswith("'--name' is a required parameter.")


def test_remove_malformed_name(script_runner):
    # Attempt removal but forget the '=' between parameter and value

    ret = script_runner.run('elyra-metadata', 'remove', METADATA_TEST_SCHEMASPACE, '--name', 'valid')
    assert ret.success is False
    assert "Parameter '--name' requires a value." in ret.stdout


def test_remove_missing(script_runner, mock_data_dir):
    # Create an instance so that the schemaspace exists.
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)
    valid = Metadata(**valid_metadata_json)
    metadata_manager.create('valid', valid)

    ret = script_runner.run('elyra-metadata', 'remove', METADATA_TEST_SCHEMASPACE, '--name=missing')
    assert ret.success is False
    assert "No such instance named 'missing' was found in the metadata-tests schemaspace." in ret.stdout

    # Now cleanup original instance.
    ret = script_runner.run('elyra-metadata', 'remove', METADATA_TEST_SCHEMASPACE, '--name=valid')
    assert ret.success


def test_remove_instance(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create('valid', valid)
    assert resource is not None
    resource = metadata_manager.create('valid2', valid)
    assert resource is not None
    another = Metadata(**another_metadata_json)
    resource = metadata_manager.create('another', another)
    assert resource is not None
    resource = metadata_manager.create('another2', another)
    assert resource is not None

    ret = script_runner.run('elyra-metadata', 'remove', METADATA_TEST_SCHEMASPACE, '--name=valid')
    assert ret.success

    ret = script_runner.run('elyra-metadata', 'remove', METADATA_TEST_SCHEMASPACE, '--name=another')
    assert ret.success

    instances = metadata_manager.get_all()
    assert len(instances) == 2
    assert instances[0].name.endswith('2')
    assert instances[1].name.endswith('2')


# Begin property tests...

def test_required(script_runner, mock_data_dir):
    # Doesn't use PropertyTester due to its unique test since all other tests require this property
    name = "required"

    expected_file = os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE, name + '.json')
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=' + name, '--display_name=' + name)

    assert ret.success is False
    assert "'--required_test' is a required parameter" in ret.stdout

    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=' + name, '--display_name=' + name,
                            '--required_test=required_value')

    assert ret.success
    assert "Metadata instance '" + name + "' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == 'metadata-test'
        assert instance_json["display_name"] == name
        assert instance_json["metadata"]["required_test"] == "required_value"


def test_number_default(script_runner, mock_data_dir):
    # Doesn't use PropertyTester due to its unique test (no failure, needs --replace, etc.)
    name = "number_default"

    expected_file = os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE, name + '.json')
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    # No negative test here.  First create w/o a value and ensure 42, then create with a value and ensure that value.
    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=' + name, '--display_name=' + name,
                            '--required_test=required_value')

    assert ret.success
    assert "Metadata instance '" + name + "' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == 'metadata-test'
        assert instance_json["display_name"] == name
        assert instance_json["metadata"]["number_default_test"] == 42

    ret = script_runner.run('elyra-metadata', 'install', METADATA_TEST_SCHEMASPACE, '--schema_name=metadata-test',
                            '--name=' + name, '--display_name=' + name,
                            '--required_test=required_value', '--replace',
                            '--number_default_test=7.2')

    assert ret.success
    assert "Metadata instance '" + name + "' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, 'metadata', METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == 'metadata-test'
        assert instance_json["display_name"] == name
        assert instance_json["metadata"]["number_default_test"] == 7.2


def test_uri(script_runner, mock_data_dir):
    prop_test = PropertyTester("uri")
    prop_test.negative_value = "//invalid-uri"
    prop_test.negative_stdout = "Property used to test uri formatting"
    #  this can be joined with previous if adding meta-properties
    #  "; title: URI Test, format: uri"
    prop_test.negative_stderr = "'//invalid-uri' is not a 'uri'"
    prop_test.positive_value = "http://localhost:31823/v1/models?version=2017-02-13"
    prop_test.run(script_runner, mock_data_dir)


def test_integer_exclusivity(script_runner, mock_data_dir):
    prop_test = PropertyTester("integer_exclusivity")
    prop_test.negative_value = 3
    prop_test.negative_stdout = "Property used to test integers with exclusivity restrictions"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Integer Exclusivity Test, exclusiveMinimum: 3, exclusiveMaximum: 10"
    prop_test.negative_stderr = "3 is less than or equal to the minimum of 3"
    prop_test.positive_value = 7
    prop_test.run(script_runner, mock_data_dir)


def test_integer_multiple(script_runner, mock_data_dir):
    prop_test = PropertyTester("integer_multiple")
    prop_test.negative_value = 32
    prop_test.negative_stdout = "Property used to test integers with multipleOf restrictions"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Integer Multiple Test, multipleOf: 6"
    prop_test.negative_stderr = "32 is not a multiple of 6"
    prop_test.positive_value = 42
    prop_test.run(script_runner, mock_data_dir)


def test_number_range(script_runner, mock_data_dir):
    prop_test = PropertyTester("number_range")
    prop_test.negative_value = 2.7
    prop_test.negative_stdout = "Property used to test numbers with range"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Number Range Test, minimum: 3, maximum: 10"
    prop_test.negative_stderr = "2.7 is less than the minimum of 3"
    prop_test.positive_value = 7.2
    prop_test.run(script_runner, mock_data_dir)


def test_const(script_runner, mock_data_dir):
    prop_test = PropertyTester("const")
    prop_test.negative_value = 2.718
    prop_test.negative_stdout = "Property used to test properties with const"
    #  this can be joined with previous if adding meta-properties
    #  " ; title: Const Test, const: 3.14"
    prop_test.negative_stderr = "3.14 was expected"
    prop_test.positive_value = 3.14
    prop_test.run(script_runner, mock_data_dir)


def test_string_length(script_runner, mock_data_dir):
    prop_test = PropertyTester("string_length")
    prop_test.negative_value = "12345678901"
    prop_test.negative_stdout = "Property used to test strings with length restrictions"
    #  this can be joined with previous if adding meta-properties
    #  "; title: String Length Test, minLength: 3, maxLength: 10"
    prop_test.negative_stderr = "'12345678901' is too long"
    prop_test.positive_value = "123456"
    prop_test.run(script_runner, mock_data_dir)


def test_string_pattern(script_runner, mock_data_dir):
    prop_test = PropertyTester("string_pattern")  # Must start/end with alphanumeric, can include '-' and '.'
    prop_test.negative_value = "-foo1"
    prop_test.negative_stdout = "Property used to test strings with pattern restrictions"
    #  this can be joined with previous if adding meta-properties
    #  "; title: String Pattern Test, pattern: ^[a-z0-9][a-z0-9-.]*[a-z0-9]$"
    prop_test.negative_stderr = "'-foo1' does not match '^[a-z0-9][a-z0-9-.]*[a-z0-9]$'"
    prop_test.positive_value = "0foo-bar.com-01"
    prop_test.run(script_runner, mock_data_dir)


def test_enum(script_runner, mock_data_dir):
    prop_test = PropertyTester("enum")
    prop_test.negative_value = "jupyter"
    prop_test.negative_stdout = "Property used to test properties with enums"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Enum Test, enum: ['elyra', 'rocks', 'added']"
    prop_test.negative_stderr = "'jupyter' is not one of ['elyra', 'rocks', 'added']"
    prop_test.positive_value = "added"
    prop_test.run(script_runner, mock_data_dir)


def test_array(script_runner, mock_data_dir):
    prop_test = PropertyTester("array")
    prop_test.negative_value = [1, 2, 2]
    prop_test.negative_stdout = "Property used to test array with item restrictions"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Array Test, minItems: 3, maxItems: 10, uniqueItems: True"
    prop_test.negative_stderr = "[1, 2, 2] has non-unique elements"
    prop_test.positive_value = [1, 2, 3, 4, 5]
    prop_test.run(script_runner, mock_data_dir)


def test_object(script_runner, mock_data_dir):
    prop_test = PropertyTester("object")
    prop_test.negative_value = {'prop1': 2, 'prop2': 3}
    prop_test.negative_stdout = "Property used to test object elements with properties restrictions"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Object Test, minProperties: 3, maxProperties: 10"
    prop_test.negative_stderr = "{'prop1': 2, 'prop2': 3} does not have enough properties"
    prop_test.positive_value = {'prop1': 2, 'prop2': 3, 'prop3': 4, 'prop4': 5}
    prop_test.run(script_runner, mock_data_dir)


def test_boolean(script_runner, mock_data_dir):
    prop_test = PropertyTester("boolean")
    prop_test.negative_value = "bogus_boolean"
    prop_test.negative_stdout = "Property used to test boolean values"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Boolean Test"
    prop_test.negative_stderr = "'bogus_boolean' is not of type 'boolean'"
    prop_test.positive_value = True
    prop_test.run(script_runner, mock_data_dir)


def test_null(script_runner, mock_data_dir):
    prop_test = PropertyTester("null")
    prop_test.negative_value = "bogus_null"
    prop_test.negative_stdout = "Property used to test null types"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Null Test"
    prop_test.negative_stderr = "'bogus_null' is not of type 'null'"
    prop_test.positive_value = None
    prop_test.run(script_runner, mock_data_dir)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import abstractmethod
from http import HTTPStatus
from logging import Logger
import os
from queue import Empty
from queue import Queue
from threading import Thread
from types import SimpleNamespace
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from jupyter_core.paths import ENV_JUPYTER_PATH
import requests
from traitlets.config import LoggingConfigurable
from traitlets.traitlets import Integer


class ComponentParameter(object):
    """
    Represents a single property for a pipeline component
    """

    def __init__(self, id: str,
                 name: str,
                 data_type: str,
                 value: str,
                 description: str,
                 required: bool = False,
                 control: str = "custom",
                 control_id: str = "StringControl",
                 items: Optional[List[str]] = None):
        """
        :param id: Unique identifier for a property
        :param name: The name of the property for display
        :param data_type: The type that the property value takes on
        :param value: The default value of the property
        :param description: A description of the property for display
        :param control: The control of the property on the display, e.g. custom or readonly
        :param control_id: The control type of the property, if the control is 'custom', e.g. StringControl, EnumControl
        :param items: For properties with a control of 'EnumControl', the items making up the enum
        :param required: Whether the property is required
        """

        if not id:
            raise ValueError("Invalid component: Missing field 'id'.")
        if not name:
            raise ValueError("Invalid component: Missing field 'name'.")

        self._ref = id
        self._name = name
        self._data_type = data_type
        self._value = value

        self._description = description
        self._control = control
        self._control_id = control_id
        self._items = items or []

        # Check description for information about 'required' parameter
        if "not optional" in description.lower() or \
                ("required" in description.lower() and
                    "not required" not in description.lower() and
                    "n't required" not in description.lower()):
            required = True

        self._required = required

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def name(self) -> str:
        return self._name

    @property
    def data_type(self) -> str:
        return self._data_type

    @property
    def value(self) -> str:
        return self._value

    @property
    def description(self) -> str:
        return self._description

    @property
    def control(self) -> str:
        return self._control

    @property
    def control_id(self) -> str:
        return self._control_id

    @property
    def items(self) -> List[str]:
        return self._items

    @property
    def required(self) -> bool:
        return bool(self._required)


class Component(object):
    """
    Represents a generic or runtime-specific component
    """

    def __init__(self, id: str, name: str,
                 description: Optional[str],
                 location_type: str,
                 location: str,
                 runtime: Optional[str] = None,
                 op: Optional[str] = None,
                 categories: Optional[List[str]] = None,
                 properties: Optional[List[ComponentParameter]] = None,
                 extensions: Optional[List[str]] = None,
                 parameter_refs: Optional[dict] = None):
        """
        :param id: Unique identifier for a component
        :param name: The name of the component for display
        :param description: The description of the component
        :param location_type: Indicates the type of component definition resource
                              location; one of ['url', filename', 'directory]
        :param location: The location of the component definition
        :param runtime: The runtime of the component (e.g. KFP or Airflow)
        :param op: The operation name of the component; used by generic components in rendering the palette
        :param categories: A list of categories that this component belongs to
        :param properties: The set of properties for the component
        :param extensions: The file extension used by the component
        """

        if not id:
            raise ValueError("Invalid component: Missing field 'id'.")
        if not name:
            raise ValueError("Invalid component: Missing field 'name'.")

        self._id = id
        self._name = name
        self._description = description
        self._location_type = location_type
        self._location = location

        self._runtime = runtime
        self._op = op
        self._categories = categories or []
        self._properties = properties

        if not parameter_refs:
            if self._location_type == "elyra":
                parameter_refs = {
                    "filehandler": "filename"
                }
            else:
                parameter_refs = {}

        if extensions and not parameter_refs.get('filehandler'):
            Component._log_warning(f"Component '{self._id}' specifies extensions '{extensions}' but \
                                   no entry in the 'parameter_ref' dictionary for 'filehandler' and \
                                   cannot participate in drag and drop functionality as a result.")

        self._extensions = extensions
        self._parameter_refs = parameter_refs

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def location_type(self) -> str:
        return self._location_type

    @property
    def location(self) -> str:
        return self._location

    @property
    def runtime(self) -> Optional[str]:
        return self._runtime

    @property
    def op(self) -> Optional[str]:
        if self._op:
            return self._op
        else:
            return self._id

    @property
    def categories(self) -> List[str]:
        return self._categories

    @property
    def properties(self) -> Optional[List[ComponentParameter]]:
        return self._properties

    @property
    def extensions(self) -> Optional[List[str]]:
        return self._extensions

    @property
    def parameter_refs(self) -> dict:
        return self._parameter_refs

    @staticmethod
    def _log_warning(msg: str, logger: Optional[Logger] = None):
        if logger:
            logger.warning(msg)
        else:
            print(f"WARNING: {msg}")


class ComponentReader(LoggingConfigurable):
    """
    Abstract class to model component_entry readers that can read components from different locations
    """
    location_type: str = None

    max_readers = Integer(3, config=True, allow_none=True,
                          help="""Sets the number of reader threads""")

    def __init__(self, file_types: List[str]):
        super().__init__()
        self.file_types = file_types

    @property
    def resource_type(self):
        """
        The RuntimePipelineProcessor accesses this property in order to
        process components on pipeline submit/export. The value must be
        one of ('filename', 'url').
        """
        return self.location_type

    @abstractmethod
    def read_component_definition(self,
                                  location: str,
                                  location_to_def: Dict[str, str]) -> Dict[str, str]:
        """
        Read an absolute location to get the contents of a component specification file

        :param location: an absolute path to the specification file to read
        :param location_to_def: a mapping of component locations to file contents

        :returns: the given 'location_to_def' object, optionally including a new
                  key-value pair if the given component location is successfully read
        """
        raise NotImplementedError()

    def read_component_definitions(self, locations: List[str]) -> Dict[str, str]:
        """
        This function starts a number of threads ('max_reader' or fewer) that read component
        definitions in parallel.

        The 'location_to_def' variable is a mapping of a component location to its content.
        As a mutable object, this dictionary provides a means to retrieve a return value for
        each thread. If a thread is able to successfully read the content of the given
        component file location, a location-to-content mapping is added to 'location_to_def'.
        """
        location_to_def = {}

        loc_q = Queue()
        for location in self.get_absolute_locations(locations):
            loc_q.put_nowait(location)

        def read_with_thread():
            """Get a location from the queue and read contents"""
            while not loc_q.empty():
                try:
                    self.log.debug("Retrieving component definition file location from queue...")
                    loc = loc_q.get(timeout=.1)
                except Empty:
                    continue

                try:
                    self.log.debug(f"Attempting read of component definition file at location '{loc}'...")
                    self.read_component_definition(loc, location_to_def)
                except Exception:
                    self.log.warning(f"Could not read component definition file at location '{loc}'. Skipping...")

                loc_q.task_done()

        # Start 'max_reader' reader threads if registry includes more than 'max_reader'
        # number of locations, else start one thread per location
        num_threads = min(loc_q.qsize(), self.max_readers)
        for i in range(num_threads):
            Thread(target=read_with_thread).start()

        # Wait for all queued locations to be processed
        loc_q.join()

        return location_to_def

    @abstractmethod
    def get_absolute_locations(self, paths: List[str]) -> List[str]:
        """
        Returns a list of absolute paths to component specification file(s)
        based on the array of potentially relative locations given
        """
        raise NotImplementedError()


class FilesystemComponentReader(ComponentReader):
    """
    Read a singular component definition from the local filesystem
    """
    location_type = 'filename'

    def determine_location(self, location_path: str) -> str:
        """
        Determines the absolute location of a given path. Error
        checking is delegated to the calling function
        """
        # Expand path to include user home if necessary
        path = os.path.expanduser(location_path)

        # Check for absolute path
        if os.path.isabs(path):
            return path

        # If path is not absolute, default to the Jupyter share location
        path = os.path.join(ENV_JUPYTER_PATH[0], 'components', path)
        return path

    def read_component_definition(self,
                                  location: str,
                                  location_to_def: Dict[str, str]) -> Dict[str, str]:
        if not os.path.exists(location):
            self.log.warning(f"Invalid location for component: {location}")
        else:
            with open(location, 'r') as f:
                location_to_def[location] = f.read()

        return location_to_def

    def get_absolute_locations(self, paths: List[str]) -> List[str]:
        absolute_paths = []
        for path in paths:
            absolute_path = self.determine_location(path)
            if not os.path.exists(absolute_path):
                self.log.warning(f"File does not exist -> {absolute_path}")
            absolute_paths.append(absolute_path)
        return absolute_paths


class DirectoryComponentReader(FilesystemComponentReader):
    """
    Read component definitions from a local directory
    """
    location_type = 'directory'

    def get_absolute_locations(self, paths: List[str]) -> List[str]:
        absolute_paths = []
        for path in paths:
            absolute_path = self.determine_location(path)
            if not os.path.exists(absolute_path):
                self.log.warning(f"Invalid directory -> {absolute_path}")
                continue

            for filename in os.listdir(absolute_path):
                if filename.endswith(tuple(self.file_types)):
                    absolute_paths.append(os.path.join(absolute_path, filename))

        return absolute_paths

    @property
    def resource_type(self):
        """
        The RuntimePipelineProcessor accesses this property in order to process
        components on pipeline submit/export. The superclass location_type is
        used because the value must be one of ('filename', 'url').
        """
        return super().location_type


class UrlComponentReader(ComponentReader):
    """
    Read a singular component definition from a url
    """
    location_type = 'url'

    def read_component_definition(self,
                                  location: str,
                                  location_to_def: Dict[str, str]) -> Dict[str, str]:
        try:
            res = requests.get(location)
        except Exception as e:
            self.log.warning(f"Failed to connect to URL for component: {location}: {str(e)}")
        else:
            if res.status_code != HTTPStatus.OK:
                self.log.warning(f"Invalid location for component: {location} (HTTP code {res.status_code})")
            else:
                location_to_def[location] = res.text

        return location_to_def

    def get_absolute_locations(self, paths: List[str]) -> List[str]:
        return paths


class GitHubComponentReader(UrlComponentReader):
    """
    Read component definitions from a github repo
    """
    location_type = 'github'

    def get_absolute_locations(self, paths: List[str]) -> List[str]:
        pass

    @property
    def resource_type(self):
        """
        The RuntimePipelineProcessor accesses this property in order to process
        components on pipeline submit/export. The superclass location_type is
        used because the value must be one of ('filename', 'url').
        """
        return super().location_type


class ComponentParser(LoggingConfigurable):  # ABC
    _component_platform = None

    @property
    def component_platform(self) -> str:
        return self._component_platform

    @property
    def file_types(self) -> List[str]:
        return self._file_types

    @abstractmethod
    def parse(self, registry_entry: SimpleNamespace) -> Optional[List[Component]]:
        """
        Parse a component definition given in the registry entry and return
        a list of fully-qualified Component objects
        """
        raise NotImplementedError()

    def get_component_id(self, location: str, name: str) -> str:
        """
        Get a unique id for a component based on its file basename and
        it's given name.
        """
        file_basename = os.path.basename(location)
        filename = os.path.splitext(file_basename)[0]
        component_name = f"{filename}_{name.replace(' ', '')}"
        return component_name

    def _format_description(self, description: str, data_type: str) -> str:
        """
        Adds parameter type information parsed from component specification
        to parameter description.
        """
        if description:
            return f"{description} (type: {data_type})"
        return f"(type: {data_type})"

    def determine_type_information(self, parsed_type: str) -> SimpleNamespace:
        """
        Takes the type information of a component parameter as parsed from the component
        specification and returns a new type that is one of several standard options.

        """
        parsed_type_lowered = parsed_type.lower()

        data_type_info: SimpleNamespace

        # Determine if this is a "container type"
        # Prefer types that occur in a clause of the form "[type] of ..." (i.e., "container" types)
        # E.g. "a dictionary of key/value pairs" will produce the type "dictionary"
        container_types = ['dictionary', 'dict', 'set', 'list', 'array', 'arr']
        for option in container_types:
            if option in parsed_type_lowered:
                data_type = option
                if data_type in ['dict', 'dictionary']:
                    data_type = "dictionary"
                elif data_type in ['list', 'set', 'array', 'arr']:
                    data_type = "list"

                # Since we know the type, create our return value and bail
                data_type_info = ComponentParser.create_data_type_info(parsed_data=parsed_type_lowered,
                                                                       data_type=data_type)
                break
        else:  # None of the container types were found...
            # Standardize type names
            if any(word in parsed_type_lowered for word in ["str", "string"]):
                data_type_info = ComponentParser.create_data_type_info(parsed_data=parsed_type_lowered,
                                                                       data_type="string")
            elif any(word in parsed_type_lowered for word in ['int', 'integer', 'number']):
                data_type_info = ComponentParser.create_data_type_info(parsed_data=parsed_type_lowered,
                                                                       data_type="number",
                                                                       control_id="NumberControl",
                                                                       default_value=0)
            elif any(word in parsed_type_lowered for word in ['bool', 'boolean']):
                data_type_info = ComponentParser.create_data_type_info(parsed_data=parsed_type_lowered,
                                                                       data_type="boolean",
                                                                       control_id="BooleanControl",
                                                                       default_value=False)
            else:  # Let this be undetermined.  Callers should check for this status and adjust
                data_type_info = ComponentParser.create_data_type_info(parsed_data=parsed_type_lowered,
                                                                       data_type="string",
                                                                       undetermined=True)

        return data_type_info

    @staticmethod
    def create_data_type_info(parsed_data: str,
                              data_type: str = 'string',
                              default_value: Any = '',
                              required: bool = True,
                              control_id: str = 'StringControl',
                              control: str = 'custom',
                              undetermined: bool = False) -> SimpleNamespace:
        """Returns a SimpleNamespace instance that contains the current state of data-type parsing.

        This method is called by ComponentParser.determine_type_information() and used by subclass
        implementations to determine the current state of parsing a data-type.

        The instance will indicate that the base ComponentParser could not determine the actual data-type
        via a `True` value in its `undetermined` attribute, in which case subclass implementations
        are advised to attempt further parsing. In such cases, the rest of the attributes of the instance
        will reflect a 'string' data type as that is the most flexible data_type and, hence, the default.
        """
        dti = SimpleNamespace(parsed_data=parsed_data,
                              data_type=data_type,
                              default_value=default_value,
                              required=required,
                              control_id=control_id,
                              control=control,
                              undetermined=undetermined)
        return dti



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import copy
import json
import os
import shutil

from jupyter_server.tests.utils import expected_http_error
from jupyter_server.utils import url_path_join
import pytest
from tornado.httpclient import HTTPClientError

from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE_ID
from elyra.tests.metadata.test_utils import byo_metadata_json
from elyra.tests.metadata.test_utils import create_json_file
from elyra.tests.metadata.test_utils import get_instance
from elyra.tests.metadata.test_utils import invalid_metadata_json
from elyra.tests.metadata.test_utils import valid_metadata_json

os.environ["METADATA_TESTING"] = "1"  # Enable metadata-tests schemaspace


async def test_bogus_schemaspace(jp_fetch, bogus_location):
    # Validate missing is not found.  Remove the bogus location to ensure its not created
    shutil.rmtree(bogus_location)
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', 'bogus', 'missing')

    assert expected_http_error(e, 400)
    assert not os.path.exists(bogus_location)


async def test_missing_instance(jp_fetch, setup_data):
    # Validate missing is not found
    name = 'missing'
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, name)
    assert expected_http_error(e, 404)


async def test_invalid_instance(jp_fetch, setup_data):
    # Validate invalid throws 404 with validation message
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'invalid')
    assert expected_http_error(e, 400)


async def test_valid_instance(jp_fetch, setup_data):
    # Ensure valid metadata can be found
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE_ID, 'valid')

    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert 'schema_name' in metadata
    assert metadata['display_name'] == 'valid metadata instance'


async def test_get_instances(jp_fetch, setup_data):
    # Ensure all valid metadata can be found
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE_ID)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE_ID]
    assert len(instances) == 2
    assert isinstance(instances, list)
    assert get_instance(instances, 'name', 'another')
    assert get_instance(instances, 'name', 'valid')


async def test_get_empty_schemaspace_instances(jp_fetch, schemaspace_location, setup_data):
    # Delete the metadata dir contents and attempt listing metadata
    shutil.rmtree(schemaspace_location)
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 0

    # Now create empty schemaspace
    os.makedirs(schemaspace_location)
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE_ID)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE_ID]
    assert len(instances) == 0


async def test_get_hierarchy_instances(jp_fetch, setup_hierarchy):
    # Ensure all valid metadata can be found
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 3
    assert isinstance(instances, list)
    assert get_instance(instances, 'name', 'byo_1')
    assert get_instance(instances, 'name', 'byo_2')
    assert get_instance(instances, 'name', 'byo_3')
    byo_3 = get_instance(instances, 'name', 'byo_3')
    assert byo_3['display_name'] == 'factory'


async def test_create_instance(jp_base_url, jp_fetch):
    """Create a simple instance - not conflicting with factory instances. """

    valid = copy.deepcopy(valid_metadata_json)
    valid['name'] = 'valid'
    body = json.dumps(valid)

    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE_ID, body=body, method='POST')
    assert r.code == 201
    assert r.headers.get('Location') == url_path_join(jp_base_url, '/elyra', 'metadata',
                                                      METADATA_TEST_SCHEMASPACE_ID, 'valid')
    metadata = json.loads(r.body.decode())
    # Add expected "extra" fields to 'valid' so whole-object comparison is satisfied.
    # These are added during the pre_save(), post_save() and post_load() hooks on the
    # MockMetadataTest class instance or when default values for missing properties are applied.
    valid['pre_property'] = valid['metadata']['required_test']
    valid['post_property'] = valid['display_name']
    valid['metadata']['number_default_test'] = 42
    assert metadata == valid


async def test_create_hierarchy_instance(jp_fetch, setup_hierarchy):
    """Attempts to create an instance from one in the hierarchy. """

    byo_instance = copy.deepcopy(byo_metadata_json)
    byo_instance['display_name'] = 'user'
    byo_instance['name'] = 'byo_2'
    body = json.dumps(byo_instance)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, body=body, method='POST')
    assert expected_http_error(e, 409)

    # Confirm the instance was not changed
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 3
    assert isinstance(instances, list)
    byo_2 = get_instance(instances, 'name', 'byo_2')
    assert byo_2['display_name'] == 'factory'


async def test_create_invalid_instance(jp_fetch):
    """Create a simple instance - not conflicting with factory instances. """

    invalid = copy.deepcopy(invalid_metadata_json)
    invalid['name'] = 'invalid'
    body = json.dumps(invalid)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, body=body, method='POST')
    assert expected_http_error(e, 400)


async def test_create_instance_missing_schema(jp_fetch, schemaspace_location):
    """Attempt to create an instance using an invalid schema """

    missing_schema = copy.deepcopy(valid_metadata_json)
    missing_schema['name'] = 'missing_schema'
    missing_schema['schema_name'] = 'missing_schema'
    missing_schema.pop('display_name')
    body = json.dumps(missing_schema)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, body=body, method='POST')
    assert expected_http_error(e, 404)

    # Ensure instance was not created.  Can't use REST here since it will correctly trigger 404
    # even though an instance was created and not removed due to failure to validate (due to
    # missing schema).  Fixed by trapping the FileNotFoundError raised due to no schema.
    assert not os.path.exists(os.path.join(schemaspace_location, 'missing_schema.json'))

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'missing_schema')
    assert expected_http_error(e, 404)


async def test_update_non_existent(jp_fetch, schemaspace_location):
    """Attempt to update a non-existent instance. """

    # Try to update a non-existent instance - 404 expected...
    valid = copy.deepcopy(valid_metadata_json)
    valid['name'] = 'valid'
    valid['metadata']['number_range_test'] = 7
    body = json.dumps(valid)

    # Update (non-existent) instance
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'valid', body=body, method='PUT')
    assert expected_http_error(e, 404)


async def test_update_instance(jp_fetch, schemaspace_location):
    """Update a simple instance. """

    # Create an instance, then update
    create_json_file(schemaspace_location, 'valid.json', valid_metadata_json)
    valid = copy.deepcopy(valid_metadata_json)
    valid['name'] = 'valid'
    valid['metadata']['number_range_test'] = 7
    body = json.dumps(valid)

    # Update instance
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE_ID, 'valid', body=body, method='PUT')
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance['metadata']['number_range_test'] == 7

    # Confirm update via jp_fetch
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'valid')
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance['metadata']['number_range_test'] == 7


async def test_invalid_update(jp_fetch, schemaspace_location):
    """Update a simple instance with invalid metadata. """

    # Create an instance, then update with invalid metadata
    create_json_file(schemaspace_location, 'update_bad_md.json', valid_metadata_json)

    # Fetch it to get the valid instance
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'update_bad_md')
    assert r.code == 200
    instance = json.loads(r.body.decode())

    # Now attempt the update with bad metadata and ensure previous still exists
    valid2 = copy.deepcopy(valid_metadata_json)
    valid2['name'] = 'valid'
    valid2['metadata']['number_range_test'] = 42
    body2 = json.dumps(valid2)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE_ID, 'update_bad_md', body=body2, method='PUT')
    assert expected_http_error(e, 400)

    # Fetch again and ensure it matches the previous instance
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE_ID, 'update_bad_md')
    assert r.code == 200
    instance2 = json.loads(r.body.decode())
    assert instance2 == instance


async def test_update_fields(jp_fetch, schemaspace_location):

    # Create an instance, then update with a new field
    create_json_file(schemaspace_location, 'update_fields.json', valid_metadata_json)
    valid = copy.deepcopy(valid_metadata_json)
    valid['metadata']['number_range_test'] = 7
    body = json.dumps(valid)

    # Update instance adding number_range_test
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'update_fields', body=body, method='PUT')
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance['metadata']['number_range_test'] == 7

    # Add a new field (per schema) and remove another -
    valid['metadata'].pop('number_range_test')
    valid['metadata']['string_length_test'] = "valid len"
    body = json.dumps(valid)

    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'update_fields', body=body, method='PUT')
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance['metadata']['string_length_test'] == "valid len"
    assert 'number_range_test' not in instance['metadata']


async def test_update_hierarchy_instance(jp_fetch, setup_hierarchy):
    """Update a simple instance - that's conflicting with factory instances. """

    # Do not name intentionally, since this is an update
    byo_instance = copy.deepcopy(byo_metadata_json)
    byo_instance['display_name'] = 'user'
    byo_instance['metadata']['number_range_test'] = 7
    body = json.dumps(byo_instance)

    # Because this is considered an update, replacement is enabled.
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'byo_2', body=body, method='PUT')
    assert r.code == 200

    # Confirm the instances and ensure byo_2 is in USER area
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    metadata = json.loads(r.body.decode())
    assert isinstance(metadata, dict)
    assert len(metadata) == 1
    instances = metadata[METADATA_TEST_SCHEMASPACE]
    assert len(instances) == 3
    assert isinstance(instances, list)
    byo_2 = get_instance(instances, 'name', 'byo_2')
    assert byo_2['schema_name'] == byo_metadata_json['schema_name']
    assert byo_2['metadata']['number_range_test'] == 7

    # Attempt to rename the resource, exception expected.
    byo_2['name'] = 'byo_2_renamed'
    body = json.dumps(byo_2)

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'byo_2', body=body, method='PUT')
    assert expected_http_error(e, 400)

    # Confirm no update occurred
    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'byo_2')
    assert r.code == 200
    instance = json.loads(r.body.decode())
    assert instance['name'] == 'byo_2'


async def test_delete_instance(jp_fetch, schemaspace_location, setup_data):
    """Create a simple instance - not conflicting with factory instances and delete it. """

    # First, attempt to delete non-existent resource, exception expected.
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'missing', method='DELETE')
    assert expected_http_error(e, 404)

    create_json_file(schemaspace_location, 'valid.json', valid_metadata_json)

    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'valid', method='DELETE')
    assert r.code == 204

    # Confirm deletion
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'valid', method='DELETE')
    assert expected_http_error(e, 404)


async def test_delete_hierarchy_instance(jp_fetch, schemaspace_location, setup_hierarchy):
    """Create a simple instance - that conflicts with factory instances and delete it only if local. """

    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'byo_2', method='DELETE')
    assert expected_http_error(e, 403)

    # create local instance, delete should succeed
    create_json_file(schemaspace_location, 'byo_2.json', byo_metadata_json)

    r = await jp_fetch('elyra', 'metadata', METADATA_TEST_SCHEMASPACE, 'byo_2', method='DELETE')
    assert r.code == 204


async def test_bogus_schema(jp_fetch):
    # Validate missing is not found

    # Remove self.request (and other 'self.' prefixes) once transition to jupyter_server occurs
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'schema', 'bogus')
    assert expected_http_error(e, 404)


async def test_missing_runtimes_schema(jp_fetch):
    # Validate missing is not found
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'schema', 'runtimes', 'missing')
    assert expected_http_error(e, 404)


async def test_get_runtimes_schemas(jp_fetch):
    # Ensure all schema for runtimes can be found
    await _get_schemaspace_schemas(jp_fetch, 'runtimes', ['kfp', 'airflow'])


async def test_get_code_snippets_schemas(jp_fetch):
    # Ensure all schema for code-snippets can be found
    await _get_schemaspace_schemas(jp_fetch, 'code-snippets', ['code-snippet'])


async def test_get_test_schemas(jp_fetch):
    # Ensure all schema for metadata_tests can be found
    await _get_schemaspace_schemas(jp_fetch, METADATA_TEST_SCHEMASPACE, ['metadata-test', 'metadata-test2'])


async def test_get_runtimes_schema(jp_fetch):
    # Ensure all schema for runtimes can be found
    await _get_schemaspace_schema(jp_fetch, 'runtimes', 'kfp')


async def test_get_code_snippets_schema(jp_fetch):
    # Ensure all schema for code-snippets can be found
    await _get_schemaspace_schema(jp_fetch, 'code-snippets', 'code-snippet')


async def test_get_test_schema(jp_fetch):
    # Ensure all schema for metadata-test can be found
    await _get_schemaspace_schema(jp_fetch, METADATA_TEST_SCHEMASPACE, 'metadata-test')


async def _get_schemaspace_schemas(jp_fetch, schemaspace, expected):
    r = await jp_fetch('elyra', 'schema', schemaspace,)
    assert r.code == 200
    schemaspace_schemas = json.loads(r.body.decode())
    assert isinstance(schemaspace_schemas, dict)
    assert len(schemaspace_schemas) == 1
    schemas = schemaspace_schemas[schemaspace]
    assert len(schemas) == len(expected)
    for expected_schema in expected:
        assert get_instance(schemas, 'name', expected_schema)


async def _get_schemaspace_schema(jp_fetch, schemaspace, expected):
    r = await jp_fetch('elyra', 'schema', schemaspace, expected)
    assert r.code == 200
    schemaspace_schema = json.loads(r.body.decode())
    assert isinstance(schemaspace_schema, dict)
    assert expected == schemaspace_schema['name']
    assert schemaspace == schemaspace_schema['schemaspace']


async def test_get_schemaspaces(jp_fetch):
    expected_schemaspaces = ['runtimes', 'code-snippets']

    r = await jp_fetch('elyra', 'schemaspace')
    assert r.code == 200
    schemaspaces = json.loads(r.body.decode('utf-8'))
    assert isinstance(schemaspaces, dict)
    assert len(schemaspaces['schemaspaces']) >= len(expected_schemaspaces)

    for expected_schemaspace in expected_schemaspaces:
        assert expected_schemaspace in schemaspaces['schemaspaces']


async def test_get_schemaspace_info(jp_fetch):
    r = await jp_fetch('elyra', 'schemaspace', METADATA_TEST_SCHEMASPACE)
    assert r.code == 200
    schemaspace_info = json.loads(r.body.decode('utf-8'))
    assert 'name' in schemaspace_info
    assert 'id' in schemaspace_info
    assert 'display_name' in schemaspace_info
    assert 'description' in schemaspace_info
    assert 'schemas' not in schemaspace_info
    assert schemaspace_info['name'] == METADATA_TEST_SCHEMASPACE
    assert schemaspace_info['id'] == METADATA_TEST_SCHEMASPACE_ID


async def test_get_missing_schemaspace_info(jp_fetch):
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch('elyra', 'schemaspace', 'missing-schemaspace')
    assert expected_http_error(e, 404)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from elyra.airflow.operator import BootscriptBuilder


def test_fail_without_cos_endpoint():
    with pytest.raises(TypeError):
        BootscriptBuilder(filename="test_notebook.ipynb",
                          cos_bucket="test_bucket",
                          cos_directory="test_directory",
                          cos_dependencies_archive="test_archive.tgz")


def test_fail_without_cos_bucket():
    with pytest.raises(TypeError):
        BootscriptBuilder(filename="test_notebook.ipynb",
                          cos_endpoint="http://testserver:32525",
                          cos_directory="test_directory",
                          cos_dependencies_archive="test_archive.tgz")


def test_fail_without_cos_directory():
    with pytest.raises(TypeError):
        BootscriptBuilder(filename="test_notebook.ipynb",
                          cos_endpoint="http://testserver:32525",
                          cos_bucket="test_bucket",
                          cos_dependencies_archive="test_archive.tgz")


def test_fail_without_cos_dependencies_archive():
    with pytest.raises(TypeError):
        BootscriptBuilder(filename="test_notebook.ipynb",
                          cos_endpoint="http://testserver:32525",
                          cos_bucket="test_bucket",
                          cos_directory="test_directory")


def test_fail_without_filename():
    with pytest.raises(TypeError):
        BootscriptBuilder(cos_endpoint="http://testserver:32525",
                          cos_bucket="test_bucket",
                          cos_directory="test_directory",
                          cos_dependencies_archive="test_archive.tgz")


def test_fail_with_empty_string_as_filename():
    with pytest.raises(ValueError) as error_info:
        BootscriptBuilder(filename="",
                          cos_endpoint="http://testserver:32525",
                          cos_bucket="test_bucket",
                          cos_directory="test_directory",
                          cos_dependencies_archive="test_archive.tgz")
    assert "You need to provide a filename for the operation." == str(error_info.value)


def test_build_cmd_with_inputs_and_outputs():
    pipeline_inputs = ['test.txt', 'test2.txt']
    pipeline_outputs = ['test3.txt', 'test4.txt']

    boot_build = BootscriptBuilder(filename="test_notebook.ipynb",
                                   cos_endpoint="http://testserver:32525",
                                   cos_bucket="test_bucket",
                                   cos_directory="test_directory",
                                   cos_dependencies_archive="test_archive.tgz",
                                   inputs=pipeline_inputs,
                                   outputs=pipeline_outputs)

    assert boot_build.inputs == pipeline_inputs
    assert boot_build.outputs == pipeline_outputs

    boot_arg_list = boot_build.container_cmd.split("--")
    for arg in boot_arg_list:
        arg_value = arg.split(" ")[1]
        if 'outputs' in arg:
            assert arg_value == f"'{';'.join(pipeline_outputs)}'"
        if 'inputs' in arg:
            assert arg_value == f"'{';'.join(pipeline_inputs)}'"



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from pathlib import Path
from urllib.parse import urlparse

from minio.api import Minio
from minio.error import BucketAlreadyExists
from minio.error import BucketAlreadyOwnedByYou
from minio.error import ResponseError
from minio.error import SignatureDoesNotMatch
from traitlets.config import LoggingConfigurable


class CosClient(LoggingConfigurable):
    client = None

    def __init__(self, config=None, endpoint=None, access_key=None, secret_key=None, bucket=None, **kwargs):
        super().__init__(**kwargs)
        if config:
            self.endpoint = urlparse(config.metadata['cos_endpoint'])
            self.access_key = config.metadata['cos_username']
            self.secret_key = config.metadata['cos_password']
            self.bucket = config.metadata['cos_bucket']
        else:
            self.endpoint = urlparse(endpoint)
            self.access_key = access_key
            self.secret_key = secret_key
            self.bucket = bucket
        # Infer secure from the endpoint's scheme.
        self.secure = self.endpoint.scheme == 'https'

        self.client = self.__initialize_object_store()

    def __initialize_object_store(self):

        # Initialize minioClient with an endpoint and access/secret keys.
        self.client = Minio(endpoint=self.endpoint.netloc,
                            access_key=self.access_key,
                            secret_key=self.secret_key,
                            secure=self.secure)

        # Make a bucket with the make_bucket API call.
        try:
            if not self.client.bucket_exists(self.bucket):
                self.client.make_bucket(self.bucket)
        except BucketAlreadyOwnedByYou as ex:
            self.log.warning("Object Storage bucket already owned by you", exc_info=True)
            raise ex from ex
        except BucketAlreadyExists as ex:
            self.log.warning("Object Storage bucket already exists", exc_info=True)
            raise ex from ex
        except ResponseError as ex:
            self.log.error("Object Storage error", exc_info=True)
            raise ex from ex
        except SignatureDoesNotMatch as ex:
            self.log.error("Incorrect Object Storage credentials supplied")
            raise ex from ex

        return self.client

    def upload_file(self, file_name, file_path):
        """
        Uploads contents from a file, located on the local filesystem at `file_path`,
        as `file_name` in object storage.
        :param file_name: Name of the file object in object storage
        :param file_path: Path on the local filesystem from which object data will be read.
        :return:
        """

        try:
            self.client.fput_object(bucket_name=self.bucket,
                                    object_name=file_name,
                                    file_path=file_path)
        except BaseException as ex:
            self.log.error('Error uploading file {} to bucket {}'.format(file_path, self.bucket), exc_info=True)
            raise ex from ex

    def upload_file_to_dir(self, dir, file_name, file_path):
        """
        Uploads contents from a file, located on the local filesystem at `file_path`,
        as `file_name` in object storage.
        :param dir: the directory where the file should be uploaded to
        :param file_name: Name of the file object in object storage
        :param file_path: Path on the local filesystem from which object data will be read.
        :return:
        """
        # elyra-320 -> always use posix path as this is targeting COS filesystem
        location = Path(os.path.join(dir, file_name))
        self.upload_file(location.as_posix(), file_path)

    def download_file(self, file_name, file_path):
        """
        Downloads and saves the object as a file in the local filesystem.
        :param file_name: Name of the file object in object storage
        :param file_path: Path on the local filesystem to which the object data will be written.
        :return:
        """
        try:
            self.client.fget_object(bucket_name=self.bucket,
                                    object_name=file_name,
                                    file_path=file_path)
        except BaseException as ex:
            self.log.error('Error reading file {} from bucket {}'.format(file_name, self.bucket), exc_info=True)
            raise ex from ex

    def download_file_from_dir(self, dir, file_name, file_path):
        """
        Downloads and saves the object as a file in the local filesystem.
        :param dir: the directory where the file is located
        :param file_name: Name of the file object in object storage
        :param file_path: Path on the local filesystem to which the object data will be written.
        :return:
        """

        self.download_file(os.path.join(dir, file_name), file_path)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json

from elyra.tests.pipeline import resources

# from jupyter_server.tests.utils import expected_http_error
# from tornado.httpclient import HTTPClientError

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources


async def test_get_components(jp_fetch):
    # Ensure all valid components can be found
    response = await jp_fetch('elyra', 'pipeline', 'components', 'local')
    assert response.code == 200
    payload = json.loads(response.body.decode())
    palette = json.loads(pkg_resources.read_text(resources, 'palette.json'))
    assert payload == palette


async def test_get_component_properties_config(jp_fetch):
    # Ensure all valid component_entry properties can be found
    response = await jp_fetch('elyra', 'pipeline', 'components', 'local', 'notebook', 'properties')
    assert response.code == 200
    payload = json.loads(response.body.decode())
    properties = json.loads(pkg_resources.read_text(resources, 'properties.json'))
    assert payload == properties



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import os
import shutil
from typing import Any
from typing import List
from typing import Optional
import uuid

from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.pipeline import Pipeline


def _read_pipeline_resource(pipeline_filename):
    root = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    pipeline_path = os.path.join(root, pipeline_filename)

    with open(pipeline_path, 'r') as f:
        pipeline_json = json.load(f)

    return pipeline_json


class NodeBase(object):
    """Represents a node of a constructed pipeline based on files in resources/node_util. """
    id: str
    name: str
    outputs: List[str]
    dependencies: List[str]
    env_vars: List[str]
    image_name: str
    fail: bool
    # from subclasses
    classifier: str
    filename: str
    pipeline_name: str  # Set during pipeline construction

    def __init__(self, name: str, num_outputs: Optional[int] = 0,
                 input_nodes: Optional[List[Any]] = None,
                 image_name: Optional[str] = None,
                 fail: Optional[bool] = False):
        self.id = str(uuid.uuid4())
        self.name = name
        self.fail = fail
        self.image_name = image_name
        self.outputs = []
        for i in range(1, num_outputs + 1):
            self.outputs.append(f"{self.name}_{i}.out")

        self.inputs = []
        self.parent_operations = []
        if input_nodes:
            for node in input_nodes:
                self.inputs.extend(node.outputs)
                self.parent_operations.append(node.id)

        self.dependencies = ['node_util/*']

    def get_operation(self) -> GenericOperation:

        self.env_vars = []
        if self.fail:  # NODE_FILENAME is required, so skip if triggering failure
            if "NODE_FILENAME" in os.environ:  # remove entry if present
                os.environ.pop("NODE_FILENAME")
        else:
            self.env_vars.append(f"NODE_FILENAME={self.filename}")

        if self.inputs:
            self.env_vars.append(f"INPUT_FILENAMES={';'.join(self.inputs)}")
        if self.outputs:
            self.env_vars.append(f"OUTPUT_FILENAMES={';'.join(self.outputs)}")

        # Convey the pipeline name
        assert self.pipeline_name is not None, "Pipeline name has not been set during construction!"
        self.env_vars.append(f"PIPELINE_NAME={self.pipeline_name}")

        # Add system-owned here with bogus or no value...
        self.env_vars.append("ELYRA_RUNTIME_ENV=bogus_runtime")

        component_parameters = {
            'filename': self.filename,
            'runtime_image': self.image_name or "NA",
            'dependencies': self.dependencies,
            'env_vars': self.env_vars,
            'inputs': self.inputs,
            'outputs': self.outputs
        }
        return GenericOperation(self.id, 'execution_node', self.name, self.classifier,
                                parent_operation_ids=self.parent_operations,
                                component_params=component_parameters)


class NotebookNode(NodeBase):
    def __init__(self, name: str, num_outputs: Optional[int] = 0,
                 input_nodes: Optional[List[Any]] = None,
                 image_name: Optional[str] = None,
                 fail: Optional[bool] = False):

        super().__init__(name, num_outputs=num_outputs, input_nodes=input_nodes, image_name=image_name, fail=fail)
        self.classifier = 'execute-notebook-node'
        self.filename = f"{self.name}.ipynb"


class PythonNode(NodeBase):
    def __init__(self, name: str, num_outputs: Optional[int] = 0,
                 input_nodes: Optional[List[Any]] = None,
                 image_name: Optional[str] = None,
                 fail: Optional[bool] = False):

        super().__init__(name, num_outputs=num_outputs, input_nodes=input_nodes, image_name=image_name, fail=fail)
        self.classifier = 'execute-python-node'
        self.filename = f"{self.name}.py"


def construct_pipeline(name: str, nodes: List[NodeBase], location: str,
                       runtime_type: Optional[str] = 'local',
                       runtime_config: Optional[str] = 'local') -> Pipeline:
    """Returns an instance of a local Pipeline consisting of each node and populates the
       specified location with the necessary files to run the pipeline from that location.
    """
    pipeline = Pipeline(str(uuid.uuid4()), name, runtime_type, runtime_config)
    for node in nodes:
        node.pipeline_name = name
        pipeline.operations[node.id] = node.get_operation()
        # copy the node file into the "working directory"
        if isinstance(node, NotebookNode):
            src_file = os.path.join(os.path.dirname(__file__), 'resources/node_util/node.ipynb')
        elif isinstance(node, PythonNode):
            src_file = os.path.join(os.path.dirname(__file__), 'resources/node_util/node.py')
        else:
            assert False, f"Invalid node type detected: {node.__class__.__name__}"

        shutil.copy(src_file, os.path.join(location, node.filename))

    # copy the node_util directory into the "working directory"
    shutil.copytree(os.path.join(os.path.dirname(__file__), 'resources/node_util'), os.path.join(location, 'node_util'))

    return pipeline
